#!/bin/bash
set -e

# Create fresh directory switch
if [ ! -d "./_opam/" ] ; then
   opam switch create . 4.14.0
fi
eval $(opam env)

# Installs the local deps
opam update
#  these commands are already idempotent :)
opam pin add -n -y coq 8.15.1
opam install -y ./coq-stdpp/coq-stdpp.opam
opam install -y ./coq-iris/coq-iris.opam
opam install -y ./coq-iris-heap-lang/coq-iris-heap-lang.opam
opam install -y ./coq-autosubst/coq-autosubst.opam
opam install -y ./coq-reloc/coq-reloc.opam

set +e
NPROCS=$(nproc 2>&1)
if [ "$?" -ne 0 ] ; then
   SYSCTL=$(sysctl -n hw.ncpu)
   if [ "$?" -ne 0 ] ; then
      PROCS=1
   else
      # probably on macOS..
      PROCS=$SYSCTL
      # in this case, the standard time command is not good enough, and we need gtime
      if [ -z "$(which gtime)" ] ; then
         echo '[diaframe-warning] Probably on macOS, and gtime was not found! For best display of timings, install gtime with `brew install gnu-time`'
      fi
   fi
else
   PROCS=$NPROCS
fi
JOBS=$(( PROCS > 2 ? PROCS - 2 : 1 ))
set -e

echo
if [[ -z "$@" ]] ; then
   ARGS=--jobs=$JOBS
   echo "Building diaframe with $JOBS jobs.."
else
   ARGS="$@"
   echo "Building diaframe, calling make with arguments: $@"
fi


# Builds diaframe
make -C diaframe $ARGS diaframe diaframe-heap-lang diaframe-reloc
# builds examples with timing info
make -C diaframe $ARGS pretty-timed TGTS="logatom-voila" TIMING_INCLUDE_MEM=0
