# Diaframe 2.0

Diaframe 2.0 is a plugin for Iris aimed at (partially) automating proofs for linearizability.

This repository contains the Coq mechanization of Diaframe 2.0.

## Building the project

Create a new `opam switch` if required (i.e. `opam switch create iris-diaframe 4.14.0`).

Make sure opam can find the Coq, Iris and ReLoC packages:

    opam repo add coq-released https://coq.inria.fr/opam/released
    opam repo add iris-dev https://gitlab.mpi-sws.org/iris/opam.git
    opam pin add -y -n git+https://gitlab.mpi-sws.org/iris/reloc#d8e93826

Next, install the necessary dependencies into your switch, and compile the diaframe package:

    opam install ./ --deps-only
    make -j8 diaframe diaframe-heap-lang

To additionally install the CoqIDE, use:

    opam install ./ --deps-only coqide

After this you should be able to use `make` to compile the various examples, which are located in [supplements/diaframe_heap_lang_examples/comparison](./supplements/diaframe_heap_lang_examples/comparison):
- `make logatom-voila`: verifies the 9 logically atomic examples from the Voila benchmark
- `make logatom-other`: verifies the simplified proofs of RDCSS and the elimination stack, as well as the new verification of the Michael Scott queue.
- `make logatom-all`: verifies all logically atomic examples above
- `make reloc`: verifies various examples provided by ReLoC using the automation of this project. The examples contain verifications that a ticket lock refines a spin lock, and that a fine-grained stack refines a course-grained stack.
- `make comparison`: verifies 21/24 examples of the Diaframe1 paper, should take about 12-20 minutes on a single core
- `make comparison-all`: verifies all examples. Beware, this takes a while (~40 minutes on a single core).

## Verifying the claims

Diaframe 2.0's main claim is that it provides powerful proof automation for verifying contextual refinement and logical atomicity.
To check this, we encourage you to:
- Browse the examples in `supplements/diaframe_heap_lang_examples/logatom/` and `supplements/diaframe_reloc/examples/`, to inspect the proof burden.
- Verify these examples, see above
- Inspect and interactively check how the procedure makes use of the various hint formats.

## Correspondence to paper

### Section 2
The proof automation for ReLoC are implemented as hints in files [supplements/diaframe_reloc/reloc_automation_base.v](./supplements/diaframe_reloc/proof_automation.v), [supplements/diaframe_reloc/reloc_auto_lob.v](./supplements/diaframe_reloc/reloc_auto_lob.v) and
[supplements/diaframe_reloc/backtracking.v](./supplements/diaframe_reloc/backtracking.v):
- Case 1 is `refines_values_abduct`, line 17 in proof_automation.v
- Case 2 is `right_execution_abduct`, line 75 in proof_automation.v
- Case 3a & 3b are `left_execution_abduct`, line 47 in proof_automation.v
- Case 3c is `rec_abduct_v2`, line 27 in reloc_auto_lob.v. It is built upon `do_lob` machinery, which is implemented in [diaframe/lib/do_lob.v](./diaframe/lib/do_lob.v). Relevant hints there are `do_lob_re_intro_only`, `do_lob_reintro_more_better` and `run_do_lob_mod`.
- Case 4 is `restore_if_right_val`, line 89 in proof_automation.v
- Case 5 is `restore_or_right_execute`, line 8 in backtracking.v.
- Case 6 is implemented in [diaframe/lib/intuitionistically.v](./diaframe/lib/intuitionistically.v)
- Case 7 is implemented in [diaframe/steps/solve_instances.v](./diaframe/steps/solve_instances.v), lines 107-138
- Case 8 is implemented in the [diaframe/steps/](./diaframe/steps/) folder

### Section 3
Section 3.2 contains a proof search procedure for logically atomic triples. We will list the relevant hints that implement this procedure:
- Case 1 is `collect_modal_wp_value`, line 544 in [diaframe/symb_exec/weakestpre.v](./diaframe/symb_exec/weakestpre.v)
- Case 2a is `abduct_from_execution`, line 534 in [diaframe/symb_exec/weakestpre.v](./diaframe/symb_exec/weakestpre.v). This instance is general enough to encapsulate logically atomic specifications.
- Case 2b is the same one, but using the `reduction_step_from` instance as specification, found on line 307 in [diaframe/symb_exec/weakestpre.v](./diaframe/symb_exec/weakestpre.v).
- Case 3a is also the same one, but now using the `pure_wp_step_exec_inst_last` instance found on line 270 in [diaframe_heap_lang/specs.v](./diaframe_heap_lang/specs.v).
- Case 3b is `rec_abduct_2` in [diaframe_heap_lang/wp_auto_lob.v](./diaframe_heap_lang/wp_auto_lob.v). It also relies on the `do_lob` machinery.
- Case 4 is implemented in [diaframe/steps/solve_instances.v](./diaframe/steps/solve_instances.v), lines 107-138
- Case 5a) is a consequence of `bi_abd_from_assumption` and `abd_from_biabd` in [diaframe/hint_search/instances_base.v](./diaframe/hint_search/instances_base.v)
- Case 5b) is implemented in [diaframe/lib/atomic.v](./diaframe/lib/atomic.v) and [diaframe/lib/greatest_fixpoint.v](./diaframe/lib/greatest_fixpoint.v). Look for `abduct_aupd_as_gfp`, `gfp_introduce_all_laterable` and `atomic_acc_abd`.
- Case 6 is implemented in the [diaframe/steps/](./diaframe/steps/) folder

### Section 4
`abduct-def` can be found in [solve_defs.v](./diaframe/solve_defs.v) as `Abduction`, where the definition is slightly generalized.

`abduct-wp-val` and `abduct-sym-ex-logatom` are both discussed in section 3, as is `abduct-wp-bind`. Recursive rules like `abduct-wand` can be found in [diaframe/hint_search/lemmas_abd.v](./diaframe/hint_search/lemmas_abd.v).

The hypothesis and contexte transformer hints can be found in [solve_defs.v](./diaframe/solve_defs.v) as `TransformHyp` and `TransformCtx`. The other hints in 4.2 and 4.3 have been discussed above.

The strategy described in 4.4 is implemented by the files in the [diaframe/steps](./diaframe/steps) folder:
- Case 1 in [diaframe/steps/introduce_var.v](./diaframe/steps/introduce_var.v)
- Case 2 in [diaframe/steps/introduce_hyp.v](./diaframe/steps/introduce_hyp.v)
- Case 3 in [diaframe/steps/solve_one.v](./diaframe/steps/solve_one.v)
- Case 4 in [diaframe/steps/solve_sep.v](./diaframe/steps/solve_sep.v)
- Case 5 in [diaframe/steps/transform.v](./diaframe/steps/transform.v)

### Section 5
#### Logical atomicity evaluation
All logical atomicity examples can be found in [supplements/diaframe_heap_lang_examples/logatom/](./supplements/diaframe_heap_lang_examples/logatom/).
#### Contextual refinement evaluation
All contextual refinement examples can be found in [supplements/diaframe_reloc/examples/](./supplements/diaframe_reloc/examples/)

## Differences between the paper and the Coq development.
We do not discuss all the features and extensibilities available in Diaframe 2.0.
For example, the abduction definition has some extra parameters: a modality and an existential quantifier.
Additionally, in the implementation, we always use generic modalities, not fancy updates, in lemmas.
These should be `ModalityStrongMono` modalities: monotone over `⊢`, and satisfy `M P ∗ Q ⊢ M (P ∗ Q)`. Essentially, they are strong indexed monads.
The names in the development might also differ from the names in the paper.

## Overview of tactics
Diaframe provides several tactics to help with proofs.

#### iStepS
The main workhorse of Diaframe. It performs a single *chunk of steps* of the automation. What a *chunk of steps* is, depends on the current goal `G`:
- `G = ∀ x, G'`: introduces `x`
- `G = U -∗ G'`: introduces `U`, decomposing it if it is a `∗` or an `∃`.
- `G = A` (with `A` an atom): applies a single abduction hint. For `A = WP e {{ Φ }}` we have two cases:
  - `e` is a value `v`: new goal is `|={⊤}=> Φ v`
  - `e = K e'` and we can find a specification `SPEC {{ L }} e' {{ U }}`: new goal is `|={⊤}=> L ∗ (∀ v, U -∗ WP (K v) {{ Φ }})`
- `G = |={E1, E2}=> ∃.. x, L ∗ G'`: perform steps of the automation until `L` has been proven, and `G'` remains.
- `G = T` (with `T` a transformer goal): apply relevant transformer hints. For example:
  - `T = ⊳ G`: introduce all laters in the context, continue with goal `G'`.
  - `T = □ G`: if the spatial context is empty, introduce the `□`, continue with goal `G`

#### iStepsS
`iStepsS` can be thought of as simply `repeat iStepS`. It runs Diaframe's automation until it gets stuck, or no goal remains.

#### iSmash
Sometimes `iStepsS` gets stuck on disjunctions, or goals of the form `|={E1, E2}=> (L1 ∨ L2) ∗ G`. In this case, you may want to just try if `L1` can be proven, and if not, try `L2`. This is what `iSmash` does: it is basically `iStepsS`, but starts backtracking on choices for the disjunction when it detects a disjunction in the goal.

### Debugging tactics
For debugging purposes, there are two additional tactics.

#### iStepDebug
This puts the goal in a format the automation can understand, but without performing any steps. It prepares the proof state for the next tactic:

#### solveStep
The `solveStep` tactic performs precisely a single step as described in the paper. This can be useful if Diaframe cannot find a hint you thought it would, or finds a different hint. In this case, you can repeat `solveStep` until it differs from your expectation. You can also use `solveStep with <HYPNAME>` to obtain two new goals. The first goal requires you to prove a hint, the second goal is the remaining goal with your hint applied.

## Extending or changing Diaframe

If you wish to extend or change Diaframe, this section provides some information to get you started.
- Important definitions can be found in [solve_defs.v](./diaframe/solve_defs.v) and
[util_classes.v](./diaframe/util_classes.v). We recommend getting familiar with them before making large changes.
- The rules and corresponding tactics for running the proof automation can all be found in [diaframe/steps](./diaframe/steps). There is a file for each relevant goal shape. If you wish to add or change the applied rules, or improve the pure automation, this is the place.
- The rules and procedure for constructing hints recursively can be found in [diaframe/hint_search](./diaframe/hint_search). There are separate files for the recursive rules themselves, and the procedure using them. This is not the appropriate folder for adding new base hints!
- If you wish to add a new hint (library), this can be done in [diaframe/lib](./diaframe/lib). You can inspect the files in that folder for examples. Note that new hint libraries dont necessarily need to be added to this folder: these can also reside in your own project!
- If you wish to add support for a new language, or experiment with changing the support for heap_lang, look at the [diaframe_heap_lang/](./diaframe_heap_lang/) folder. This folder contains all files responsible for making Diaframe understand `heap_lang` verification goals.
- The infrastructure responsible for symbolic execution can be found in [diaframe/symb_exec](./diaframe/symb_exec). This is set up quite abstractly, and is capable of handling goals other that WP. Proper documentation of this part is still future work.

## Directory structure

The Coq files reside in one of the following 5 top-level directories:
- [diaframe](./diaframe/): contains Diaframe's proof search procedure, not instantiated for a language or goal
- [diaframe_heap_lang](./diaframe_heap_lang/): contains an instantiation of Diaframe for weakest-preconditions on `heap_lang`, Iris's default language
- [tutorial](./tutorial): contains two tutorial files with lots of comments, to get users started on Diaframe
- [supplements](./supplements/): contains verifications of various examples, and instantiations of Diaframe for other projects: ReLoC, Simuliris, Actris, λ-rust, iris-examples
- [tests](./tests/): contains some files to test Diaframe's proof search, used for development

### Files in [diaframe](./diaframe/)
The following files contain definitions and utilities used throughout the project
- [solve_defs.v](./diaframe/solve_defs.v): Defines the hint formats `BiAbd` and `Abduct` and the goal formats `SolveSep` and `SolveOne`
- [util_classes.v](./diaframe/util_classes.v): Defines several typeclasses used throughout the project. An important ones is `ModalityEC`. Roughly speaking, if `ModalityEC M` is true, and we can `SplitModality3 M ?M1 ?M2` the proof-search strategy works behind modality `M`.
- [util_instances.v](./diaframe/util_instances.v): Contains instances of typeclasses defined above
- [env_utils.v](./diaframe/env_utils.v): Contains some additional lemmas for manipulating IPM's environments
- [tele_utils.v](./diaframe/tele_utils.v): Contains machinery and lemmas to be able to work under existentials, and to be able to separate relevant from irrelevant existentials.
- [proofmode_base.v](./diaframe/proofmode_base.v): Bundles Diaframe's basic proof search strategy in one file. Importing this will give you access to the automation tactics, but the available hints are (intentionally) minimal.

This folder also contains the following subfolders:

#### Files in [diaframe/hint_search](./diaframe/hint_search)
These files are all related to (bi-)abduction hints
- [lemmas_biabd.v](./diaframe/hint_search/lemmas_biabd.v): Contains proofs of all recursive rules for constructing bi-abduction hints.
- [lemmas_abd.v](./diaframe/hint_search/lemmas_abd.v): Contains proofs of all recursive rules for constructing abduction hints.
- [search_biabd.v](./diaframe/hint_search/search_biabd.v): Defines the search procedure for bi-abduction hints, using the recursive rules.
- [search_abd.v](./diaframe/hint_search/search_abd.v): Defines the search procedure for abduction hints, using the recursive rules.
- [instances_base.v](./diaframe/hint_search/instances_base.v): Defines the base hint instances necessary for minimum functionality.
- [explicit_hint_registration.v](./diaframe/hint_search/explicit_hint_registration.v): Defines a way to reset the registered hints, so the number of distinct hints used can be counted.

#### Files in [diaframe/steps](./diaframe/steps)
These files contain the proofs and tactics for executing the proof-search strategy.
- [solve_sep.v](./diaframe/steps/solve_sep.v): Contains lemmas and tactics for making progress on `Δ ⊢ || M || ∃ x, A ∗ G` goals (`SolveSep` goals). In particular, contains the proof of the application of bi-abduction hints.
- [solve_one.v](./diaframe/steps/solve_one.v): Contains lemmas and tactics for making progress on `Δ ⊢  M (∃ x, A)` goals (`SolveOne` goals). In particular,
contains the proof of the application of abduction hints.
- [introduce_hyp.v](./diaframe/steps/introduce_hyp.v): Contains lemmas and tactis for making progress on `Δ ⊢ U -∗ G` goals (`IntroduceHyp` goals). This file is responsible for decomposing `U`s into `H_C`s, and for finding interference between hypotheses (i.e. `MergablePersist` or `MergableConsume` instances).
- [introduce_var.v](./diaframe/steps/introduce_var.v): Contains lemmas and tactics for making progress on `Δ ⊢ ∀ x. G x` goals (`IntroduceVars` goals). Quite short, no special things happen here.
- [transform.v](./diaframe/steps/transform.v): Contains lemmas and tactics for making progress on goals like `Δ ⊢ ⊳ G` (`Transform` goals). Responsible for introducing various modalities, as well as Löb induction and introducing atomic updates.
- [solve_instances.v](./diaframe/steps/solve_instances.v): Contains some additional typeclass instances necessary for using Diaframe's automation.
- [pure_solver.v](./diaframe/steps/pure_solver.v): Defines the pure solver used in various places throughout the project.
- [small_steps.v](./diaframe/steps/small_steps.v): Combines the lemmas and tactics from this folder to construct the `solveSteps` tactic, which runs the proof-search strategy as long as the goals are `SolveSep`, `SolveOne`, `IntroduceHyp`, `IntroduceVars` or `Transform`. In this way, it ensures we can run the proof-search strategy to execute a single statement, without observing all tiny steps in between. This is useful for debugging and using the strategy in interactive mode.
- [tactics.v](./diaframe/steps/tactics.v): Uses the `solveSteps` tactic from `small_steps.v` to build the `iStepS` and `iStepsS` tactics. `iStepS` turns the goal into one of `SolveSep`, `SolveOne`, `IntroduceHyp`, `IntroduceVars` or `Transform`, then runs `solveSteps`. `iStepsS` roughly repeatedly calls `iStepS`, and thus runs the proof-search strategy continuously. This file contains some additional instances that make raw `solveStep` goals work better in interactive Iris proofs.
- [verify_tac.v](./diaframe/steps/verify_tac.v): Uses `iStepsS` to build `verify_tac`, a version of `iStepsS` tailored to being run with `Program`.

The files below are also in this folder, but less relevant:
- [local_post.v](./diaframe/steps/local_post.v): a tactic which can be used to add a form of ''proof outlines'' to an Iris proof. Used in the `barrier` and `peterson` example, mainly useful for reducing number of goals to prove if there are a lot of disjunctions.
- [namer_tacs.v](./diaframe/steps/namer_tacs.v): an interface which can be used to provide better names for automatically introduced hypotheses.

#### Files in [diaframe/symb_exec](./diaframe/symb_exec)
This folder contains some infrastructure for generic assertions that can ''execute'', like WP. This is not discussed in the paper.
- [defs.v](./diaframe/symb_exec/defs.v): contains definitions of a generic symbolic execution interface
- [weakestpre.v](./diaframe/symb_exec/weakestpre.v): contains an instantiation of this for WP.
- [spec_notation.v](./diaframe/symb_exec/spec_notation.v): defines the `SPEC` notation, in terms of a `ReductionStep` defined in `defs.v`. Comes with an `unfold_spec_goal`, which unfolds a `SPEC` goals into one using just `WP`s.
- [weakestpre_logatom.v](./diaframe/symb_exec/weakestpre_logatom.v): extends the WP symbolic execution instances to support symbolic execution with logically atomic specifications.
- [atom_spec_notation.v](./diaframe/symb_exec/atom_spec_notation.v): defines more `SPEC` notation for logically atomic triples.

#### Files in [diaframe/lib](./diaframe/lib)
This folder defines several ghost-state and hint libraries used in different examples.
- [ticket_cmra.v](./diaframe/lib/ticket_cmra.v): basically a wrapper around the `CoPset` cmra, that separates singletons and the set of all numbers below/under a given one. Used in the `ticket_lock` and the `barrier` examples.
- [frac_token.v](./diaframe/lib/frac_token.v): Defines the `token`, `no_tokens` and `token_counter` resources from the ARC example. Actually defined on top of:
- [tokenizable.v](./diaframe/lib/tokenizable.v): More abstract version of above, does not constrain `token`s to only work for Fractionals.
- [Z_cmra.v](./diaframe/lib/Z_cmra.v): Instantiates the integers `Z` with addition as a cmra. The `≼` operation is trivial on these.
- [int_as_nat_diff.v](./diaframe/lib/int_as_nat_diff.v): It is sometimes useful to represent `Z` as the difference between two (rising) natural numbers. This file extends the pure automation with such a concept. Used in `rwlock_linux_bounded` and the `inc_dec` counter.
- [except_zero.v](./diaframe/lib/except_zero.v): Makes Diaframe strip off the except zero modality when possible when introducing hypotheses. Additionally, makes Diaframe strip off laters of Timeless hypotheses.
- [persistently.v](./diaframe/lib/persistently.v): Adds support for proving goals behind a `<pers>` modality.
- [intuitionistically.v](./diaframe/lib/intuitionistically.v): Adds support for proving goals behind a `□` modality.
- [iris_hints.v](./diaframe/lib/iris_hints.v): Adds hints specific to Iris's logic - rules for manipulating invariants.
- [own_hints.v](./diaframe/lib/own_hints.v): Adds hints for working with `own γ a` hypotheses, where `a` is a ghost element built from Iris's cmra building blocks (like `auth`, `agree`, `prod`, `option`, `excl`, etcetera.). This also relies on files in the [lib/own](./diaframe/lib/own/) folder.
- [atomic.v](./diaframe/lib/atomic.v): Adds hints to deal with proving and using atomic updates (`AU`s)
- [laterable.v](./diaframe/lib/laterable.v): Adds hints to deal with the make_laterable modality
- [do_lob.v](./diaframe/lib/do_lob.v): Adds hints to perform Löb induction automatically. This file is not focused at Löb induction on a particular goal: as soon as it sees a `do_lob` goal, the hints will gather and generalize over relevant hypotheses and variables, perform Löb induction, then present the final goal as `post_lob G = G`. This `post_lob` 'modality' can then be used as a syntactic indicator to run additional rules, such as unfolding recursive definitions for `WP`s.
- [greatest_laterable_fixpoint.v](./diaframe/lib/greatest_laterable_fixpoint.v): used by [atomic.v](./diaframe/lib/atomic.v), defines rules for introducing a greatest 'laterable' fixpoint, ie, a greatest fixpoint, but where all hypotheses must be laterable before introducing the fixpoint. An atomic update can be seen as the greatest laterable fixpoint of an atomic access (`AACC` in Iris notation).

#### Files in [diaframe/legacy](./diaframe/legacy)
Can be ignored, contain old experiments.

___
### Files in [diaframe_heap_lang](./diaframe_heap_lang/)
Contains all files relevant for instantiating the proof-search strategy for HeapLang
- [base_hints](./diaframe_heap_lang/base_hints.v): registers hints related to HeapLang resources, mainly `l ↦ v`. Additionally strengthens the pure solver to solve (in)equalities on values.
- [specs](./diaframe_heap_lang/specs.v): registers specifications of HeapLangs primitive operations
- [atomic_specs](./diaframe_heap_lang/atomic_specs.v): registers (logically atomic) specification for HeapLangs `atomic_heap` abstraction
- [proof_automation.v](./diaframe_heap_lang/proof_automation.v): bundles all required imports for the proof-search strategy instantiated for HeapLang, into one file.
- [loc_map.v](./diaframe_heap_lang/loc_map.v): It is sometimes useful to be able to couple locations to specific data (often `gname`s). This file makes that doable without much hassle -- `loc_map` is a heap-lang specific hint library.
- [wp_auto_lob.v](./diaframe_heap_lang/wp_auto_lob.v): If this file is imported, Diaframe will automatically perform Löb induction on weakestpreconditions of recursive functions.
- [examples/base/benchmark.v](./diaframe_heap_lang/examples/base/benchmark.v): contains some basic examples Diaframe should be able to handle.
- [examples/base/list_functions.v](./diaframe_heap_lang/examples/base/list_functions.v): contains some basic (sequential) examples Diaframe should be able to handle. Mainly tests some Löb induction things.
- [examples/base/swap.v](./diaframe_heap_lang/examples/base/swap.v): contains a really basic swap function Diaframe should be able to verify.

### Folders in [supplements](./supplements/)
The [supplements](./supplements/) folder contains additional instantiations of the Diaframe proof search procedure, as well as verifications of various examples. It also contains the `.opam` files with build dependencies.

#### Files in [supplements/diaframe_heap_lang_examples](./supplements/diaframe_heap_lang_examples)
Contains 3 folders:
- [comparison](./supplements/diaframe_heap_lang_examples/comparison): Contains the 24 examples discussed in the PLDI '22 paper.
- [logatom](./supplements/diaframe_heap_lang_examples/logatom): Showcases the proof-search strategy on verifications of logical atomicity, as discussed in the OOPSLA '23 draft paper. One of the examples is the novel verification of logical atomicity of the Michael-Scott queue.
- [wip](./supplements/diaframe_heap_lang_examples/wip): Contains various examples that are still work-in-progress.

#### Files in [supplements/diaframe_reloc](./supplements/diaframe_reloc)
Contains an [examples](./supplements/diaframe_reloc/examples/) folder with ReLoC examples, as well as the following files, responsible for Diaframe's instantiation for ReLoC:
- [tp_specs](./supplements/diaframe_reloc/tp_specs.v): adds specifications for the right-hand side expressions in refinement judgments.
- [symb_exec](./supplements/diaframe_reloc/symb_exec.v): instantiates Diaframe's [symb_exec](./diaframe/symb_exec) infrastructure for ReLoC's judgments.
- [proof_automation](./supplements/diaframe_reloc/proof_automation.v): bundles all required files to give useful proof automation for ReLoC
- [reloc_auto_lob](./supplements/diaframe_reloc/reloc_auto_lob.v): adds hints to perform automatic Löb induction for ReLoC.
- [atomic_post_update](./supplements/diaframe_reloc/atomic_post_update.v): adds a slightly changed version of atomic updates (`AU`s) that dont fix the closing mask beforehand. This is necessary to allow symbolic execution of right-hand side expressions after using a logically atomic spec for the left-hand side.
- [symb_exec_logatom](./supplements/diaframe_reloc/symb_exec_logatom.v): actually makes ReLoC's symbolic execution compatible with logically atomic specifications for left-hand side expressions.
