From iris.proofmode Require Import base coq_tactics reduction tactics.
From diaframe Require Import util_classes tele_utils solve_defs.
From iris.bi Require Import bi telescopes.

Import bi.

(* This file contains all base (bi)abduction hints, needed for minimal functionality of Diaframe.
*)

Set Universe Polymorphism.

Section biabd_instances.
  Context {PROP : bi}.

  Implicit Types P : PROP.
  Implicit Types M : PROP → PROP.

  Global Instance bi_abd_from_assumption p P P' :
    (* this requirement will remember newly proven persistent things *)
    TCIf (TCAnd (TCEq p false) $ Persistent P) (TCEq P' P) (TCEq P' emp%I) →
    HINT □⟨p⟩ P ✱ [- ; emp] ⊫ [id]; P ✱ [P'] | 50.
  Proof.
    rewrite /BiAbd /= !right_id //.
    case => [[-> HP] -> /= | ->].
    - apply persistent_entails_r => //.
    - rewrite right_id.
      apply intuitionistically_if_elim.
  Qed.

  Global Instance bi_abd_from_assumption_intuit P :
    HINT □⟨true⟩ P ✱ [- ; emp] ⊫ [id]; □ P ✱ [emp].
  Proof. by rewrite /BiAbd /= !right_id. Qed.

  Global Instance bi_abd_empty_goal_from_emp p :
    HINT □⟨p⟩ emp ✱ [- ; emp] ⊫ [@id PROP]; χ ✱ [emp] | 50.
  Proof. by rewrite /BiAbd /= empty_goal_eq bi.intuitionistically_if_elim. Qed.


  Class NormalizePropWrap (P Q : PROP) :=
    normalize_prop_wrap : NormalizeProp P Q true.

  Global Instance wrap_through_eq P Q Q' : 
    TCNoBackTrack (NormalizeProp P Q' true) →
    TCEq Q Q' →
    NormalizePropWrap P Q.
  Proof. case. by move => HPQ ->. Qed.

  (* Whether this is hint is definitely required is debatable. but it is useful *)
  Global Instance normalize_disj (TT : tele) (P P1 P2 P' : TT -t> PROP) teq :
    (TC∀.. (tt : TT), FromOr (tele_app P tt) (tele_app P1 tt) (tele_app P2 tt)) →
    SimplTeleEq TT teq →
    (TC∀.. (tt : TT), NormalizePropWrap (tele_app P tt) (tele_app P' tt)) →
    BiAbd (TTl := TT) (TTr := TT) false (ε₀)%I P id P' (tele_map (tele_map (bi_pure)) (teq))%I | 40.
  Proof.
    rewrite /NormalizePropWrap /NormalizeProp /BiAbd /SimplTeleEq /= empty_hyp_first_eq => _ /tforall_forall Hteq /tforall_forall HP.
    apply tforall_forall => ttl.
    rewrite left_id.
    iIntros "HP".
    iExists (ttl).
    rewrite HP !tele_map_app.
    iFrame; iPureIntro.
    specialize (Hteq ttl).
    apply (dep_eval_tele ttl) in Hteq.
    apply Hteq.
    apply tele_eq_app_both.
  Qed.

  Global Instance abd_if_bool_decide `{Decision φ} (L R : PROP) :
    (* cannot be a biabd, since the forall would not be introducable! *)
    HINT1 ε₀ ✱ [∀ (b : bool), (⌜b = true⌝ ∗ ⌜φ⌝ ∨ ⌜b = false⌝ ∗ ⌜¬φ⌝ )-∗ if b then L else R] 
        ⊫ [id]; if bool_decide φ then L else R.
  Proof.
    rewrite /Abduct /=.
    case_bool_decide.
    - rewrite (forall_elim true) empty_hyp_first_eq left_id.
      iIntros "H".
      iApply "H". by iLeft.
    - rewrite (forall_elim false) empty_hyp_first_eq left_id.
      iIntros "H".
      iApply "H". by iRight.
  Qed.

  Global Instance biabd_emp_valid {TTl TTr} p P Q M R S :
    AsEmpValid (BiAbd (TTl := TTl) (TTr := TTr) p P Q M R S) (∀.. (ttl : TTl), □?p P ∗ (tele_app R ttl) -∗ M $ ∃.. (ttr : TTr), tele_app Q ttr ∗ tele_app (tele_app S ttl) ttr).
  Proof.
    split.
    - rewrite /BiAbd /= => /tforall_forall HPQ.
      iIntros (ttl) "HPR". rewrite HPQ //.
    - rewrite /BiAbd tforall_forall /= => HPQ ttl.
      iApply HPQ.
  Qed.
End biabd_instances.

Section abduction_instances.
  Context {PROP : bi}.
  Implicit Types M : PROP → PROP.

  Global Instance abd_from_biabd {TTl TTr : tele} p P Q M R S :
    BiAbd (TTl := TTl) (TTr := TTr) p P Q M R S →
    ModalityMono M →
    (TC∀.. (ttr : TTr), TCOr (∀.. ttl : TTl, Affine (tele_app (tele_app S ttl) ttr)) (Absorbing (tele_app Q ttr))) →
    Abduct (TT := TTr) p P Q M (∃.. ttl, tele_app R ttl)%I. 
    (* priority of this instance has a significant performance impact! I think low = better *)
  Proof.
    rewrite /BiAbd /Abduct => /tforall_forall HPR HM /tforall_forall Hdrop.
    apply wand_elim_r', bi_texist_elim => ttl.
    apply wand_intro_l.
    rewrite HPR.
    apply HM, bi_texist_mono => ttr.
    case: (Hdrop ttr) => [/(dep_eval_tele ttl) HS | HQ ].
    - by rewrite HS right_id.
    - by apply sep_elim_l, _.
  Qed.

  Global Instance abduct_emp_valid {TT} p P Q M R :
    AsEmpValid (Abduct (TT := TT) p P Q M R) (□?p P ∗ R -∗ M $ ∃.. (tt : TT), tele_app Q tt)%I.
  Proof.
    split; rewrite /Abduct /=.
    - move => -> //. iIntros "$".
    - move => HPQ. iApply HPQ.
  Qed.
End abduction_instances.

Unset Universe Polymorphism.









































