From iris.proofmode Require Import base coq_tactics reduction tactics.
From diaframe Require Import util_classes tele_utils solve_defs util_instances.
From diaframe.hint_search Require Import lemmas_abd search_biabd.
From iris.bi Require Import bi telescopes.

Import bi.

(* This file uses the recursive rules for constructing abduction hints 
   to build a procedure which finds hints using these rules.
*)

Set Universe Polymorphism.

Section witness_postpone_optimization.
  Context {PROP : bi}.
  Implicit Types (M : PROP → PROP).

  Lemma abd_rec_mod_forall_unif p P' `(P : A → PROP) {C : tele} TT Q M M1 M2 Mh R (g : C -t> A) :
    IntoForallCareful P' P →
    SplitLeftModality3 M M1 M2 →
    (∀ (c : C), ∃ a' R',
      AbductModRec (TT := TT) p (P a') Q M2 Mh R' ∧
      TCAnd (GatherEvarsEq a' (tele_app g c)) $ (* if g : C → A maybe the tactic can be simpler..? *)
      TCEq R' (R c)
    ) →
    IntroducableModality Mh →
    ModalityMono Mh →
    AbductModRec (TT := TT) p P' Q M M1 (∃.. c, R c).
  Proof.
    rewrite /IntoForallCareful => HP HMs Habd HMh HMh'.
    (* need to do some work before we can use abd_forall *)
    rewrite /AbductModRec. etrans.
    apply HMs. rewrite HP.
    assert ((∃.. c, R c) ⊢ (∃.. c, tele_app (tele_bind R) c)) as ->.
    { apply bi_texist_mono => tt. by rewrite tele_app_bind. }
    done.
    eapply (abd_rec_mod_forall) => // c.
    destruct (Habd c) as [a' [R' [HPR [Ha' HR]]]].
    revert Ha' HPR => /eq_from_gather_evars_eq ->.
    rewrite /AbductModRec => <-. apply HMh'. 
    apply bi.sep_mono_r. rewrite tele_app_bind HR //.
  Qed.

  Lemma abd_rec_mod_witness_unif {A : Type} {C : tele} {TTf : A → tele} (g : C -t> A) M M1 M2 Mh R P (Q : TeleS TTf -t> PROP) p :
    SplitLeftModality3 M M1 M2 →
    (∀ (c : C), ∃ a' R''',
      AbductModRec (TT := TTf a') p P (Q a') M2 Mh R''' ∧ (* R''': PROP! so stuff is not as difficult as in BiAbd *)
      TCAnd (GatherEvarsEq a' (tele_app g c)) $
      TCEq (R''') (R c)
    ) →
    IntroducableModality Mh →
    AbductModRec (TT := TeleS TTf) p P Q M M1 (∃.. c, R c).
  Proof.
    move => HMs HPR HM.
    rewrite /AbductModRec.  etrans.
    apply HMs.
    assert ((∃.. c, R c) ⊢ (∃.. c, tele_app (tele_bind R) c)) as -> => //.
    { apply bi_texist_mono => tt. by rewrite tele_app_bind. }
    eapply abd_rec_mod_goal_exist => //.
    move => c.
    destruct (HPR c) as [a' [R' [HPR2 [Ha' ->]]]].
    revert Ha' HPR2 => /eq_from_gather_evars_eq ->.
    rewrite /AbductModRec => HPR2.
    erewrite <-HPR2. rewrite tele_app_bind //.
  Qed.

End witness_postpone_optimization.

Section stages_rec_mod.
  Context {PROP : bi}.
  Implicit Types (M : PROP → PROP).

  Definition AbductGoalModRec {PROP : bi} := @AbductModRec PROP.

  Lemma abd_goal_mod_rec_from_base p P {TTr} Q M M1 M2 S :
    TCAnd (Abduct (TT := TTr) p P Q M2 S) (SplitModality3 M M1 M2) →
    AbductGoalModRec TTr p P Q M M1 S.
  Proof.
    case => HPQ HMs.
    change @AbductGoalModRec with @AbductModRec.
    eapply abduct_mod_rec_from_abduct => //.
    apply HMs. apply HMs.
  Qed.

  Section hyp.
    Lemma abdhyp_rec_mod_mod_intro_l p P M1 P' {TTr} Q M2 S M Mh Ml :
      IntoModal p P M1 P' →
      SplitModality4 M Ml M1 M2 → (* it is valid and sometimes desired to keep a part of Ml around *)
      AbductModRec (TT := TTr) false P' Q M2 Mh S →
      ModalityStrongMono Mh →
      AbductModRec (TT := TTr) p P Q M Ml (Mh S).
    Proof.
      intros.
      eapply (abd_modrec_id_swap _ _ _ _ _ (M1 ∘ M2)); try apply H0.
      eapply abd_rec_mod_mod_intro_l; try done; try apply H0.
      move => ? //.
    Qed.
  End hyp.

End stages_rec_mod.


Ltac find_abd_mod_input_initial := 
  notypeclasses refine (abd_mod_input_from_mod_rec _ _ _ _ _ _ _ _).

Ltac abd_mod_rec_later_mono cont :=
  simple eapply abd_rec_mod_later_mono; [iSolveTC .. | cont | ].

Ltac abd_mod_rec_sepl cont := eapply abd_rec_mod_sepl; [iSolveTC| cont | ].
Ltac abd_mod_rec_sepr cont := eapply abd_rec_mod_sepr; [iSolveTC| cont | ].
Ltac abd_mod_rec_sep P cont := 
  assert_succeeds (assert (∃ P1 P2, IntoSepCareful P P1 P2) as _ by (eexists; eexists; iSolveTC)); 
    ((abd_mod_rec_sepl cont) + (abd_mod_rec_sepr cont)).

Ltac abd_mod_rec_hyp_exist cont := 
  simple eapply abd_rec_mod_exist; [iSolveTC | 
        cbn [tforall tele_fold tele_bind tele_app tele_appd_dep]; intros; cont | ..].

Ltac abd_mod_rec_hyp_mod cont := 
  eapply abdhyp_rec_mod_mod_intro_l; [iSolveTC | iSolveTC | cont | ..].

Ltac abd_mod_rec_wand cont := 
  eapply abd_rec_mod_wand; [iSolveTC | cont | ..].

Ltac abd_mod_rec_forall cont := 
  eapply abd_rec_mod_forall_unif; [iSolveTC | iSolveTC | 
      intros ?; do 2 eexists; split; [cont | ] 
      | | ].


Ltac abd_mod_rec_from_goal :=
  change @AbductModRec with @AbductGoalModRec.

Ltac abd_mod_rec_atom_shortcut p P :=
  assert_succeeds (assert (AtomAndConnective p P) by iSolveTC);
  abd_mod_rec_from_goal.

Ltac abd_mod_rec_step_hyp_naive p P cont := 
  (abd_mod_rec_atom_shortcut p P; cont) + (
  (abd_mod_rec_later_mono cont) ||
  (abd_mod_rec_sep P cont) ||
  (abd_mod_rec_hyp_exist cont) ||
  (abd_mod_rec_hyp_mod cont) ||
  (abd_mod_rec_wand cont) ||
  (abd_mod_rec_forall cont) ||
  (abd_mod_rec_from_goal; cont)
  ).

Ltac abd_connective_as_atom_shortcut := fail.
(* this keeps open the option of seeing, for example, all wands also as atoms.
  Redefine as:
Ltac abd_connective_as_atom_shortcut ::= abd_mod_rec_atom_shortcut.
  to find more hints *)

Ltac abd_mod_rec_step_hyp p P cont :=
  (abd_connective_as_atom_shortcut p P; cont) +
  lazymatch P with
       (* The (||) is to support some tricky laterable stuff. These ensure we don't quit too early when 
          looking for hints of the shape  Abduct _ (▷ P) Q _ _ _  -> so no matching later in goal *)
  | bi_later _ => (abd_mod_rec_later_mono cont) || (abd_mod_rec_from_goal; cont)
  | bi_laterN _ _ => (abd_mod_rec_later_mono cont) || (abd_mod_rec_from_goal; cont)
  | bi_sep _ _ => abd_mod_rec_sep P cont
  | bi_exist _ => abd_mod_rec_hyp_exist cont
  | fupd _ _ _ => abd_mod_rec_hyp_mod cont
  | bupd _ => abd_mod_rec_hyp_mod cont
  | bi_wand _ _ => abd_mod_rec_wand cont
  | bi_forall _ => abd_mod_rec_forall cont
  | _ => abd_mod_rec_step_hyp_naive p P cont
  end.


Ltac abd_mod_rec_goal_mod := 
  eapply abd_mod_later_r; [iSolveTC.. | ].

Ltac abd_mod_rec_goal_exist cont :=
  eapply abd_rec_mod_witness_unif; [iSolveTC | 
      intros ?; do 2 (notypeclasses refine (ex_intro _ _ _)); split; [cont | ] 
    | ].

Ltac abd_mod_rec_step_goal cont :=
  (solve [eapply abd_goal_mod_rec_from_base; diaframe_hint_search]) || (
    tryif abd_mod_rec_goal_mod then cont else
    abd_mod_rec_goal_exist cont
  ).

Ltac abd_mod_rec_step_with cont :=
  lazymatch goal with
  | |- AbductModRec ?p ?P _ _ _ _ => abd_mod_rec_step_hyp p P cont
  | |- AbductGoalModRec _ _ _ _ _ _ _ => abd_mod_rec_step_goal cont
  end.

Ltac abd_mod_rec_step_test := abd_mod_rec_step_with ltac:(idtac).

Ltac find_abd_mod_rec_steps :=
  let cont := ltac:(pre_step; find_abd_mod_rec_steps) in
  abd_mod_rec_step_with cont.

Ltac find_abd_mod_input := find_abd_mod_input_initial; [find_abd_mod_rec_steps| ]; solve [iSolveTC].

Global Hint Extern 4 (AbductModInput _ _ _ _ _ _) =>
  find_abd_mod_input : typeclass_instances.








