From iris.algebra Require Export cmra local_updates proofmode_classes.
From iris.prelude Require Import options.

(* Z with add as the operation. This is like nat, but '≼' becomes trivial *)
Section ZR.
  Local Instance Z_valid_instance : Valid Z := λ x, True.
  Local Instance Z_validN_instance : ValidN Z := λ n x, True.
  Local Instance Z_pcore_instance : PCore Z := λ x, Some 0%Z.
  Local Instance Z_op_instance : Op Z := Z.add.
  Definition z_op (x y : Z) : x ⋅ y = (x + y)%Z := eq_refl.
  Lemma Z_included (x y : Z) : x ≼ y ↔ True.
  Proof. 
    split => // _. 
    exists (y - x)%Z.
    rewrite z_op; fold_leibniz; lia. 
  Qed.
  Lemma Z_ra_mixin : RAMixin Z.
  Proof.
    apply ra_total_mixin; try by eauto.
    - solve_proper.
    - intros x y z. apply Z.add_assoc.
    - intros x y. apply Z.add_comm.
    - by exists 0.
  Qed.
  Canonical Structure ZR : cmra := discreteR Z Z_ra_mixin.

  Global Instance Z_cmra_discrete : CmraDiscrete ZR.
  Proof. apply discrete_cmra_discrete. Qed.

  Local Instance Z_unit_instance : Unit Z := 0.
  Lemma Z_ucmra_mixin : UcmraMixin Z.
  Proof. split; apply _ || done. Qed.
  Canonical Structure ZUR : ucmra := Ucmra Z Z_ucmra_mixin.

  Global Instance Z_cancelable (x : Z) : Cancelable x.
  Proof. by intros ???? ?%Z.add_cancel_l. Qed.

  Lemma Z_local_update (x y x' y' : Z) :
    (x + y')%Z = (x' + y)%Z → (x,y) ~l~> (x',y').
  Proof.
    intros ??; apply local_update_unital_discrete=> z _.
    rewrite !z_op. fold_leibniz => Hx. split => //. lia.
  Qed.

  (* This one has a higher precedence than [is_op_op] so we get a [+] instead
     of an [⋅]. *)
  Global Instance Z_is_op (z1 z2 : Z) : IsOp (z1 + z2)%Z z1 z2.
  Proof. done. Qed.
End ZR.

From iris.base_logic Require Import invariants.
From iris.proofmode Require Import base classes reduction tactics.
From diaframe.lib Require Import own_hints.

Section ZR_automation.
  Implicit Types Σ : gFunctors.

  Global Instance Z_find_local_update (x x' y : Z) :
    FindLocalUpdate x x' y (x' + y - x)%Z True | 150.
  Proof.
    move => _.
    apply Z_local_update.
    lia.
  Qed.

  Global Instance Z_unit : AsCmraUnit (A := ZUR) 0%Z.
  Proof. done. Qed.

  Global Instance nat_valid_op (a a1 a2 : Z) Σ: 
    IsOp a a1 a2 → IsValidOp a a1 a2 Σ True%I.
  Proof. apply from_isop. Qed.

  Global Instance ucmra_subtract_Z (z1 z2 : Z) : 
    UcmraSubtract z1 z2 True (z1 - z2)%Z.
  Proof. rewrite /UcmraSubtract /CmraSubtract /= z_op => _. fold_leibniz. lia. Qed.

  Global Instance Z_included_merge (z1 z2 : Z) {Σ} : IsIncludedMerge z1 z2 (Σ := Σ) True%I.
  Proof.
    rewrite /IsIncludedMerge.
    iIntros "_"; iSplit. 
    - eauto.
    - iIntros "%". iPureIntro => /=. exists (z2 - z1)%Z. fold_leibniz. rewrite z_op. lia.
  Qed.

End ZR_automation.


















