From diaframe Require Import proofmode_base.
From diaframe.hint_search Require Import search_biabd.
From iris.bi Require Import bi.
From iris.proofmode Require Import proofmode.
From iris.base_logic.lib Require Import later_credits iprop fancy_updates.


Section later_credits_merging.
  Context `{!lcGS Σ}.

  Global Instance merge_later_credits (n m : nat) : 
    MergableConsume (PROP := iPropI Σ) (£ n) (λ p Pin Pout,
    TCAnd (TCEq Pin (£ m)) (TCEq Pout (£ (n + m)))).
  Proof.
    move => p Pin Pout [-> ->].
    rewrite bi.intuitionistically_if_elim.
    by rewrite lc_split.
  Qed.

  Global Instance prove_later_credits (n m m' : nat) :
    nat_cancel.NatCancel m n O m' →
    HINT £ n ✱ [- ; emp] ⊫ [id]; £ m ✱ [£ m'] | 60.
  Proof.
    rewrite /nat_cancel.NatCancel => Hnm; iStepS.
    by rewrite -lc_split -Hnm lc_split.
  Qed.
End later_credits_merging.


Section gathered_laters.
  Context {PROP : bi}.
  Implicit Types P : PROP.

  Definition gathered_later_def n P : PROP := ▷^n P.

  Definition gathered_later_aux : seal (@gathered_later_def). Proof. by eexists. Qed.
  Definition gathered_later : nat → PROP → PROP := gathered_later_aux.(unseal).
  Lemma gathered_later_eq : @gathered_later = @gathered_later_def.
  Proof. by rewrite -gathered_later_aux.(seal_eq). Qed.

  Definition gathered_later_go_aux : seal (@gathered_later_def). Proof. by eexists. Qed.
  Definition gathered_later_go : nat → PROP → PROP := gathered_later_go_aux.(unseal).
  Lemma gathered_later_go_eq : @gathered_later_go = @gathered_later_def.
  Proof. by rewrite -gathered_later_go_aux.(seal_eq). Qed.

  Definition start_gathering_def (n : nat) : PROP := emp.
  Definition start_gathering_aux : seal (@start_gathering_def). Proof. by eexists. Qed.
  Definition start_gathering : nat → PROP := start_gathering_aux.(unseal).
  Lemma start_gathering_eq : @start_gathering = @start_gathering_def.
  Proof. by rewrite -start_gathering_aux.(seal_eq). Qed.

  Global Instance gathered_later_hint n :
    HINT ε₀ ✱ [P ; start_gathering n ∗ gathered_later_go n P] ⊫ [id] Q; gathered_later n Q ✱ [⌜P = Q⌝].
  Proof.
    do 2 iStepS. rewrite gathered_later_go_eq gathered_later_eq /gathered_later_def.
    iExists x. by iFrame.
  Qed.

  Global Instance start_gathering_hint n :
    HINT ε₀ ✱ [- ; emp] ⊫ [id]; start_gathering n ✱ [gathered_later_go n True].
  Proof.
    iStepsS.
    rewrite start_gathering_eq gathered_later_go_eq /gathered_later_def.
    iStepsS.
  Qed.

  Global Instance gathered_later_go_merge_spatial P Q Q' n m m' : MergableConsume (gathered_later_go n P) (λ p Pin Pout,
    TCAnd (IntoLaterNMax Pin m Q) $
    TCAnd (TCOr (TCAnd (SolveSepSideCondition (m ≤ n)) $ TCAnd (TCEq Q Q') (TCEq m' O)) (TCAnd (nat_cancel.NatCancel m n m' O) (TCEq Q' (▷^m' Q)%I))) $
           TCEq Pout (gathered_later_go n (P ∗ □?p Q'))).
  Proof.
    rewrite /SolveSepSideCondition /nat_cancel.NatCancel => p Pin Pout [-> [[[Hmn [-> Hm']] | [Hm' ->]] ->]];
    rewrite gathered_later_go_eq /gathered_later_def.
    - assert (Q' ⊢ ▷^(n - m) Q') as HQ' by iStepsS. rewrite {1}HQ'.
      rewrite -bi.laterN_add.
      replace (m + (n - m)) with n by lia.
      rewrite {1}bi.laterN_intuitionistically_if_2.
      iStepsS.
    - replace m with (n + m') by lia. clear Hm'.
      rewrite bi.laterN_add.
      rewrite {1}bi.laterN_intuitionistically_if_2.
      iStepsS.
  Qed.
End gathered_laters.


Section later_recursive_hints.
  Context `{!invGS Σ}.

  (* lemmas for n-ary credit elimination *)
  Lemma lc_fupd_elim_laterN E P n :
     £ n -∗ (▷^n P) -∗ |={E}=> P.
  Proof.
    revert P.
    elim: n; first eauto.
    move => n IH P /=.
    rewrite lc_succ.
    iIntros "[H£ H£n] HP".
    iMod (lc_fupd_elim_later with "H£ HP") as "HP".
    iPoseProof (IH with "H£n HP") as "$".
  Qed.

  Lemma lc_fupd_add_later n E1 E2 P :
    (* below could probably be generalized to the step taking fupd.. *)
    £ n -∗ (▷^n |={E1, E2}=> P) -∗ |={E1, E2}=> P.
  Proof.
    iIntros "Hf Hupd". iApply (fupd_trans E1 E1).
    iApply (lc_fupd_elim_laterN with "Hf Hupd").
  Qed.

  (* We get one additional hint with later credits: eliminate a later with a credit 
      to inspect whats beneath. *)
  Lemma biabd_later_credit_elim p P n P' {TTl TTr : tele} Q R S M E1 E2 :
    IntoLaterNMax P n P' → (* we can't use IntoLaterN since the n is an input there, not an output. *)
    BiAbd (TTl := TTl) (TTr := TTr) p P' Q M R S →
    TCOr (TCEq M (fupd E1 E2)) (TCAnd (TCEq M id) (TCEq E1 E2)) → (* TODO: this is brittle. phrase using le_upd modality? *)
    BiAbd (TTl := TTl) (TTr := TTr) p P Q (fupd E1 E2) (tele_map (flip bi_sep (£ n)) R) S.
  Proof.
    rewrite /IntoLaterNMax /BiAbd /= !tforall_forall => HPP HPR HM ttl.
    rewrite tele_map_app /= HPP bi.laterN_intuitionistically_if_2.
    iIntros "(HP & HR & H£)".
    iMod (lc_fupd_elim_laterN with "H£ HP") as "HP".
    iStopProof. rewrite comm HPR.
    case: HM => [-> | [-> ->] /=]; first done.
    apply fupd_intro.
  Qed.

  Set Universe Polymorphism.

  (* actually, we can do more: it is okay to prove R behind a later, and we can strip all hyps that are currently later *)
  Lemma biabd_later_credit_elim_adv p P n P' {TTl TTr : tele} Q R S M E1 E2 :
    IntoLaterNMax P n P' → 
    BiAbd (TTl := TTl) (TTr := TTr) p P' Q M R S →
    TCOr (TCEq M (fupd E1 E2)) (TCAnd (TCEq M id) (TCEq E1 E2)) → 
    BiAbd (TTl := TeleS (λ _ : iProp Σ, TTl)) (TTr := TTr) p P Q (fupd E1 E2) (λ lP, tele_map (flip bi_sep (£ n ∗ gathered_later n lP)%I) $ tele_map (bi_laterN n) R) 
      (λ lP, tele_map (tele_map (flip bi_sep lP)) S).
  Proof.
    rewrite /IntoLaterNMax /BiAbd => HPP /tforall_forall HPR HM V.
    apply tforall_forall => ttl /=.
    rewrite !tele_map_app /= HPP bi.laterN_intuitionistically_if_2.
    iIntros "(HP & HR & H£ & HV)".
    iCombine "HP HR" as "HPR". rewrite HPR gathered_later_eq /gathered_later_def.
    setoid_rewrite tele_map_app => /=.
    (* I think to strip laters from additional things in our context, we need to do it as in this lemma.
       we cannot change the S with something like a forall, since we need to know what resources to strip laters from NOW *)
    iMod (lc_fupd_elim_laterN with "H£ [HV HPR]") as "HPV". by iCombine "HV HPR" as "HPV".
    iStopProof. rewrite comm.
    case: HM => [-> | [-> ->] /=].
    - rewrite modality_strong_frame_l.
      apply fupd_mono, bi.wand_elim_l', bi_texist_elim => ttr.
      rewrite !bi_texist_exist -(bi.exist_intro ttr).
      apply bi.wand_intro_r. by rewrite assoc.
    - apply bi.wand_elim_l', bi_texist_elim => ttr.
      rewrite !bi_texist_exist -(bi.exist_intro ttr) -fupd_intro.
      apply bi.wand_intro_r. by rewrite assoc.
  Qed.

  Unset Universe Polymorphism.

  (* now stated as a lemma for the hint search procedure *)
  Lemma biabdhyp_later_credit_elim p P n P' {TTl TTr : tele} Q R S M M1 M1' M2 E1 E2 :
    IntoLaterNMax P n P' → (* we can't use IntoLaterN since the n is an input there, not an output. *)
    BiAbdUnfoldHypMod (TTl := TTl) (TTr := TTr) p P' Q M M1 M2 R S →
    TCOr (TCAnd (TCEq M2 (fupd E1 E2)) (TCEq M1 M1')) (TCAnd (TCEq M2 id) (TCAnd (TCEq E1 E2) (SplitModality3 M M1' (fupd E1 E2)))) →
    BiAbdUnfoldHypMod (TTl := TTl) (TTr := TTr) p P Q M M1' (fupd E1 E2) (tele_map (flip bi_sep (£ n)) R) S.
  Proof.
    move => ? [[? ?]] HM. eapply unfoldhyp_from_base => //. 
    eapply biabd_later_credit_elim => //.
    case: HM; case => HM1; last case; intros; try iSolveTC.
    case: HM => [[/TCEq_eq HM' /TCEq_eq HM''] | [/TCEq_eq HM' [/TCEq_eq HM'' HM''']]]; subst; eauto.
  Qed.

  Set Universe Polymorphism.

  Lemma biabdhyp_later_credit_elim_adv p P n P' {TTl TTr : tele} Q R S M M1 M1' M2 E1 E2 :
    IntoLaterNMax P n P' → (* we can't use IntoLaterN since the n is an input there, not an output. *)
    BiAbdUnfoldHypMod (TTl := TTl) (TTr := TTr) p P' Q M M1 M2 R S →
    TCOr (TCAnd (TCEq M2 (fupd E1 E2)) (TCEq M1 M1')) (TCAnd (TCEq M2 id) (TCAnd (TCEq E1 E2) (SplitModality3 M M1' (fupd E1 E2)))) →
    BiAbdUnfoldHypMod (TTl := TeleS (λ _ : iProp Σ, TTl)) (TTr := TTr) p P Q M M1' (fupd E1 E2) (λ lP, tele_map (flip bi_sep (£ n ∗ gathered_later n lP)%I) $ tele_map (bi_laterN n) R)
      (λ lP, tele_map (tele_map (flip bi_sep lP)) S).
  Proof.
    move => ? [[? ?]] HM. eapply unfoldhyp_from_base => //. 
    eapply biabd_later_credit_elim_adv => //.
    case: HM; case => HM1; last case; intros; try iSolveTC.
    case: HM => [[/TCEq_eq HM' /TCEq_eq HM''] | [/TCEq_eq HM' [/TCEq_eq HM'' HM''']]]; subst; eauto.
  Qed.

  Unset Universe Polymorphism.
End later_recursive_hints.


(* tactic which applies the recursive rule *)
Ltac biabd_step_later_credit_elim cont :=
(*  notypeclasses refine (biabdhyp_later_credit_elim _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _); [iSolveTC | cont | ].*)
  notypeclasses refine (biabdhyp_later_credit_elim_adv _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _); [iSolveTC | cont | ].


(* and override search to actually use it *)
Ltac biabd_step_hyp_naive p P cont ::=
  (biabd_step_atom_shortcut; cont) + (
  tryif biabd_step_later_mono then cont else (* ensures ⊳ (P ∗ Q) is not also tried as ⊳P ∗ ⊳ Q *)
  tryif biabd_step_sep P then cont else
  (biabd_step_later_credit_elim cont) ||
  (biabd_step_hyp_exist cont) || (*TODO rewrite this so that the single tactics like <- make progress/don't skip if a hint could not be found *)
  (biabd_step_hyp_mod cont) ||
  (biabd_step_wand; cont) ||
  (biabd_step_forall cont) ||
  (biabd_step_from_goal p P; cont)
  ).

Ltac biabd_step_hyp p P cont ::=
  (connective_as_atom_shortcut; cont) +
  lazymatch P with
       (* we have tryifs here to support some tricky Laterable stuff. These ensure we don't quit too early when 
          looking for hints of the shape  BiAbd _ (▷ P) Q _ _ _  -> so no matching later in goal *)
  | bi_later _ => tryif biabd_step_later_mono then cont else 
        ((biabd_step_later_credit_elim cont) || (biabd_step_from_goal p P; cont))
  | bi_laterN _ _ => tryif biabd_step_later_mono then cont else
        ((biabd_step_later_credit_elim cont) || (biabd_step_from_goal p P; cont))
  | bi_sep _ _ => biabd_step_sep P; cont
  | bi_exist _ => biabd_step_hyp_exist cont
  | fupd _ _ _ => biabd_step_hyp_mod cont
  | bupd _ => biabd_step_hyp_mod cont
  | bi_wand _ _ => biabd_step_wand; cont
  | bi_forall _ => biabd_step_forall cont
  | _ => biabd_step_hyp_naive p P cont
  end.

Section test.
  Context `{!invGS Σ}.

  Lemma test P Q : ▷ (P ∗ Q) ∗ £ 1 ⊢ |={⊤}=> P ∗ Q.
  Proof. iStepsS. Qed.
End test.




