From iris.proofmode Require Import base coq_tactics reduction tactics.
From diaframe Require Import util_classes util_instances solve_defs tele_utils.
From iris.bi Require Import bi telescopes.

Export solve_defs.

(* This file contains the minimum instances necessary for functionality of the automation.
   Concretely, it provides some instances responsible for getting appropriate
   SolveOne/IntroduceHyp/IntroduceVars/Transform goals from a regular Iris goal. *)

(* TODO: remove this, superfluous, since SolveOne will do it one step later.
   Its current only use is to present a readable goal if iStepS fails while not making any progress *)
Class FromTExistAlways {PROP : bi} (P : PROP) (TT : tele) (P' : TT -t> PROP) :=
  from_texist_always : (∃.. tt, tele_app P' tt) ⊢ P.


Section as_solve_goal_instances.
  Context {PROP : bi}.
  Implicit Types (M : PROP → PROP).

  Global Instance fromtexist_always_fromtexist (P : PROP) TT P' :
    FromTExist P TT P' →
    FromTExistAlways P TT P' | 5.
  Proof. done. Qed.

  Global Instance fromtexists_refl (P : PROP) :
    FromTExistAlways P [tele] P | 10.
  Proof. by rewrite /FromTExistAlways /=. Qed.

  (* By default, we only see connectives as solve goals when we have native support for them in the hint search.
    So those are: wand, forall, later, and the default ∗, possibly behind a modality *)


  Global Instance fromtexist_into_solve_goal M P TT P' :
    ModalityMono M →
    FromTExistAlways P TT P' →
    AsSolveGoal M P (SolveOne M P') | 99.
  Proof.
    unseal_diaframe; rewrite /AsSolveGoal /FromTExistAlways => HM HP'.
    by apply HM.
  Qed.

  (* Wand has preference over forall, since FromForall will trigger on goals of the form (⌜φ⌝ -∗ R) -> (∀ (_ : φ), R) *) 

  Global Instance mod_wand_as_solve_goal M P H G :
    IntroducableModality M →
    FromWand P H G →
    AsSolveGoal M P (IntroduceHyp H G) | 38.
  Proof. 
    unseal_diaframe; rewrite /AsSolveGoal /FromWand /= /ModWeaker /ModalityMono.
    move => <- /= -> //.
  Qed.

  Global Instance mod_later_wand_as_solve_goal M P P' n H G :
    IntroducableModality M →
    FromLaterNMax P n P' →
    FromWand P' H G →
    AsSolveGoal M P (Transform $ (bi_laterN n) $ IntroduceHyp H G) | 39. (* FromWand does not look under laters, FromForall does... *)
  Proof. 
    unseal_diaframe; rewrite /AsSolveGoal /FromLaterNMax /FromWand /= /ModWeaker.
    move => <- /= <- <- //.
  Qed.

  Global Instance mod_forall_as_solve_goal M P A (Φ : A → PROP) ident :
    IntroducableModality M → 
    FromForall P Φ ident →
    AsSolveGoal M P (IntroduceVars (TT := [tele_pair A]) Φ) | 40.
  Proof. 
    unseal_diaframe; rewrite /AsSolveGoal /FromForall /= /ModWeaker.
    move => /= <- -> //.
  Qed.

  Global Instance mod_later_forall_as_solve_goal M P P' n A (Φ : A → PROP) ident :
    IntroducableModality M →
    FromLaterNMax P n P' →
    FromForall P' Φ ident →
    AsSolveGoal M P (Transform $ (bi_laterN n) $ IntroduceVars (TT := [tele_pair A]) Φ) | 50.
  Proof. 
    unseal_diaframe; rewrite /AsSolveGoal /FromLaterNMax /FromForall /= /ModWeaker.
    move => <- /= <- <- //.
  Qed.

  Global Instance later_as_solve_goal (P : PROP) :
    AsSolveGoal id (▷ P)%I (Transform $ (bi_later) P).
  Proof. unseal_diaframe. rewrite /AsSolveGoal //. Qed.

  Global Instance laterN_as_solve_goal (P : PROP) n :
    AsSolveGoal id (▷^n P)%I (Transform $ (bi_laterN n) P).
  Proof. unseal_diaframe. rewrite /AsSolveGoal //. Qed.
End as_solve_goal_instances.


Section as_solve_goal_continue_instances.
  Context {PROP : bi}.
  Implicit Types (M : PROP → PROP).

  Global Instance solve_sep_as_solve_goal {TT} M P R :
    AsSolveGoal id (SolveSep (TT := TT) M P R) (SolveSep (TT := TT) M P R) | 20.
  Proof. unseal_diaframe; rewrite /AsSolveGoal //. Qed.

  Global Instance solve_one_as_solve_goal {TT} M P :
    AsSolveGoal id (SolveOne (TT := TT) M P) (SolveOne (TT := TT) M P) | 20.
  Proof. unseal_diaframe; rewrite /AsSolveGoal //. Qed.
End as_solve_goal_continue_instances.


Section later_transform.
  Context {PROP : bi}.
  Implicit Types P : PROP.
  (* Since ▷ is part of the grammar and of the recursive hint rules, 
    we always include the ▷ transform rules *)

  Global Instance introducable_in_if_always M Δ P :
    IntroducableModality M → TRANSFORM Δ, M P =[ctx]=> P | 10.
  Proof.
    rewrite /TransformCtx => HM.
    rewrite envs_entails_unseal => -> //.
    apply HM.
  Qed.

  Global Instance transform_later_hyp p P P' R : 
    IntoLaterN false 1 P P' →
    TRANSFORM □⟨p⟩ P, ▷ R =[hyp]=> (▷ IntroduceHyp (□?p P') R).
  Proof.
    rewrite /TransformHyp; unseal_diaframe => ->. destruct p; simpl;
    iIntros "HQR HQ !>";
    by iApply "HQR".
  Qed.

  Global Instance transform_laterN_hyp p n P P' R : 
    IntoLaterN false n P P' →
    TRANSFORM □⟨p⟩ P, ▷^n R =[hyp]=> (▷^n IntroduceHyp (□?p P') R).
  Proof.
    rewrite /TransformHyp; unseal_diaframe => ->. destruct p; simpl;
    iIntros "HQR HQ !>";
    by iApply "HQR".
  Qed.
End later_transform.


Section collect_modality_instances.
  Context {PROP : bi}.
  (* These instances ensure CollectModal looks deep enough behind nested modalities *)

  Global Instance collect_modal_prepend (B1 B2 : PROP) M :
    PrependModality B1 M B2 →
    CollectModal B1 M B2 | 90.
  Proof. rewrite /CollectModal /PrependModality // => -> //. Qed.

  Global Instance modal_compose_bupd_fupd E1 E2 (P : PROP) `{BiBUpdFUpd PROP}  :
    CollectModal (|==> |={E1,E2}=> P) (fupd E1 E2) P | 45.
  Proof.
    rewrite /CollectModal.
    by iIntros "$".
  Qed.

  Global Instance modal_compose_fupd_bupd E1 E2 (P : PROP) `{BiBUpdFUpd PROP}  :
    CollectModal (|={E1,E2}=> |==> P) (fupd E1 E2) P | 45.
  Proof.
    rewrite /CollectModal.
    by iIntros ">$".
  Qed.

  Global Instance modal_compose_fupd_fupd_l2 E1 E2 E3 (P : PROP) `{BiFUpd PROP}  :
    CollectModal (|={E3}=> |={E1,E2}=> P)%I (fupd E1 E2) P | 45.
  Proof.
    rewrite /CollectModal.
    by iIntros "$".
  Qed.

  Global Instance modal_compose_bupd_bupd (P : PROP) `{BiBUpd PROP } :
    CollectModal (|==> |==> P) bupd P | 45.
  Proof.
    rewrite /CollectModal.
    by iIntros "$".
  Qed.

  Lemma collect_modal_modal (P : PROP) M :
    ModalityStrongMono M →
    CollectModal (M P) M P.
  Proof. rewrite /CollectModal //. Qed.

  Global Instance collect_modal_bupd (P : PROP) `{BiBUpd PROP } :
    CollectModal (|==> P) bupd P | 50.
  Proof. apply: collect_modal_modal. Qed.

  Global Instance collect_modal_fupd (P : PROP) `{BiFUpd PROP } E1 E2 :
    CollectModal (|={E1,E2}=> P) (fupd E1 E2) P | 50.
  Proof. apply: collect_modal_modal. Qed.
End collect_modality_instances.


Section drop_modal_instances.
  Context {PROP : bi}.
  Implicit Types R : PROP.

  (* this causes better inspection on things like □ ⋄ ⌜φ⌝ *)
  Global Instance drop_modal_persistently R:
    DropModal (bi_intuitionistically)%I R (λ P, TCAnd (Persistent P) (Affine P)).
  Proof.
    rewrite /DropModal => P HP.
    iIntros "HPR #HP".
    iApply "HPR". iApply "HP".
  Qed.

  Global Instance drop_modal_affine R :
    DropModal (bi_affinely) R (λ P, Affine P).
  Proof.
    rewrite /DropModal => P HP.
    iIntros "HPR HP".
    by iApply "HPR".
  Qed.

  Global Instance IntoModal_affinely p R : 
    IntoModal p (<affine> R) bi_affinely R | 30.
  Proof. rewrite /IntoModal bi.intuitionistically_if_elim //. Qed.
End drop_modal_instances.


Section misc.
  Context {PROP : bi}.
  (* TODO: where to put this? *)

  Global Instance exclusive_prop_merge_false Q : 
    ExclusiveProp Q →
    MergableConsume (PROP := PROP) Q (λ p Pin Pout,
      TCAnd (TCEq Pin Q) $
            TCEq Pout False%I).
  Proof.
    rewrite /ExclusiveProp /MergableConsume => HQ p Pin Pout [-> ->].
    rewrite bi.intuitionistically_if_elim HQ //.
  Qed.
End misc.

