From iris.proofmode Require Import base coq_tactics reduction tactics.
From diaframe Require Import util_classes solve_defs tele_utils steps.namer_tacs.
From iris.bi Require Import bi telescopes.

Import bi.

(* This file contains 'case 4' of Diaframe's proof search strategy: making progress on atoms behind modalities.
   Lone atoms cause Diaframe to look for an abduction hint.
*)

Section solve_one_rules.

  Context {PROP : bi}.
  Implicit Types M : PROP → PROP.

  Lemma solve_one_fromsep {TT : tele} M P P1 P2 {TT1} P1' {TT2} h M1 M2 :
    (TC∀.. (tt : TT), FromSepCareful (tele_app P tt) (tele_app P1 tt) (tele_app P2 tt)) →
    ExtractDep P1 TT1 P1' TT2 h →
    SplitLeftModality3 M M1 M2 →
    SolveSep M1 P1' $ 
      tele_bind (λ tt1, M2 $ bi_texist $ tele_app $ tele_fun_compose' P2 (tele_app h tt1))
    ⊢ SolveOne M P.
  Proof.
    unseal_diaframe; rewrite /TCTForall /FromSepCareful /ExtractDep !tforall_forall => HP HP1 [HM HM1 HM2].
    rewrite -HM => {HM}.
    apply HM1 => {HM1}.
    apply bi_texist_elim => tt1.
    move: HP1 => /(dep_eval tt1) /tforall_forall HP1.
    rewrite tele_app_bind modality_strong_frame_r.
    apply modality_strong_is_mono => {HM2}.
    apply wand_elim_l', bi_texist_elim => tt2.
    apply wand_intro_l.
    rewrite !tele_fun_compose_eq'.
    erewrite HP1, HP.
    by rewrite bi_texist_exist -(exist_intro _).
  Qed.

  Lemma solve_one_fromexist_nested {TT1 : tele} {TT2} M P P' :
    (∀.. (tt : TT1), FromTExist (tele_app P tt) (tele_app TT2 tt) (tele_appd_dep P' tt)) →
    ModalityMono M →
    SolveOne (TT := TeleConcatType TT1 TT2) M (tele_curry P') 
    ⊢ SolveOne (TT := TT1) M P.
  Proof.
    unseal_diaframe => /tforall_forall HPP' HM /=.
    apply HM, bi_texist_elim => tt.
    rewrite (tele_arg_concat_inv_eq tt).
    destruct (tele_arg_concat_inv_strong_hd tt) as [tt1 tt2].
    rewrite bi_texist_exist -(exist_intro tt1).
    rewrite -HPP' !bi_texist_exist.
    rewrite -(exist_intro tt2).
    rewrite tele_uncurry_curry_compose.
    by rewrite -tele_curry_uncurry_relation.
  Qed.

  Lemma solve_one_not_introducable {TT'} M Q M1 M2 :
    NotIntroducable M →
    SplitLeftModality3 M M1 M2 →
    SolveSep (TT := [tele]) M1 empty_goal (M2 (∃.. tt, tele_app Q tt)) ⊢
    SolveOne (TT := TT') M Q.
  Proof.
    unseal_diaframe => _ [HM HM' HM''].
    rewrite -HM /=.
    apply HM'.
    rewrite empty_goal_eq left_id //.
  Qed.

  Lemma solve_one_abd {TT} p P M Q R M1 M2 :
    Abduct (TT := TT) p P Q M2 R →
    SplitModality3 M M1 M2 →
    □?p P ∗ M1 R ⊢ SolveOne (TT := TT) M Q.
  Proof.
    unseal_diaframe; rewrite /Abduct => HPR [HM HM1 HM2].
    rewrite -HM modality_strong_frame_r.
    apply HM1 => {HM1} /=.
    by rewrite comm HPR.
  Qed.

  Lemma solve_one_abd_mod_input {TT} p P M Q R :
    AbductModInput TT p P Q M R →
    □?p P ∗ R ⊢ SolveOne (TT := TT) M Q.
  Proof.
    unseal_diaframe => /=. 
    rewrite /AbductModInput /Abduct => HPR //.
  Qed.
End solve_one_rules.


Section coq_tactics.
  Context {PROP : bi}.
  Implicit Types M : PROP → PROP.

  Lemma tac_solve_one_fromsep {TT : tele} Δ M M1 M2 P P1 P2 {TT1} P1' {TT2} h :
    (TC∀.. (tt : TT), FromSepCareful (tele_app P tt) (tele_app P1 tt) (tele_app P2 tt)) →
    ExtractDep P1 TT1 P1' TT2 h →
    SplitLeftModality3 M M1 M2 →
    envs_entails Δ $ SolveSep M1 P1' $ 
      tele_bind (λ tt1, M2 $ bi_texist $ tele_app $ tele_fun_compose' P2 (tele_app h tt1)) →
    envs_entails Δ $ SolveOne (TT := TT) M P.
  Proof.
    rewrite envs_entails_unseal => HP1P2 HP1P1' HM ->.
    by eapply solve_one_fromsep.
  Qed.

  Lemma tac_solve_one_fromexist_nested {TT1 : tele} Δ {TT2} M P P' :
    (TC∀.. (tt : TT1), FromTExist (tele_app P tt) (tele_app TT2 tt) (tele_appd_dep P' tt)) →
    ModalityMono M →
    envs_entails Δ $ SolveOne (TT := TeleConcatType TT1 TT2) M (tele_curry P') →
    envs_entails Δ $ SolveOne (TT := TT1) M P.
  Proof.
    rewrite envs_entails_unseal => HPP' HM ->.
    by apply solve_one_fromexist_nested.
  Qed.

  Lemma tac_solve_one_ext_abd_input Δ i M {TT} Q R p P :
    FindInExtendedContext Δ (λ p P, AbductModInput TT p P Q M R) i p P →
    envs_entails (envs_option_delete false i p Δ) R →
    envs_entails Δ $ SolveOne (TT := TT) M Q.
  Proof.
    rewrite envs_entails_unseal => HΔiP.
    rewrite (findinextcontext_spec false) => ->.
    apply ficext_satisfies in HΔiP as HPQ.
    rewrite <-solve_one_abd_mod_input => //.
  Qed.

  Lemma tac_solve_one_pure (Δ : envs PROP) {TT : tele} b M P φ :
    (TC∀.. tt : TT, FromPure b (tele_app P tt) (tele_app φ tt)) →
    IntroducableModality M →
    (∃.. tt, tele_app φ tt) →
    (TCIf (TCEq b true) (TCForall Affine (env_spatial Δ)) (TCEq b false)) → 
    envs_entails Δ $ SolveOne (TT := TT) M P.
  Proof.
    rewrite envs_entails_unseal texist_exist => /tforall_forall HPφ HM [tt Hφ].
    unseal_diaframe; rewrite -HM /= bi_texist_exist -(exist_intro tt) -HPφ /= pure_True //=.
    case => [-> HΔ | ->] /=; eauto.
  Qed.

  Lemma tac_solve_one_not_introducable {TT'} Δ M Q M1 M2 :
    NotIntroducable M →
    SplitLeftModality3 M M1 M2 →
    envs_entails Δ $ SolveSep (TT := [tele]) M1 empty_goal (M2 (∃.. tt, tele_app Q tt))%I →
    envs_entails Δ $ SolveOne (TT := TT') M Q.
  Proof.
    rewrite envs_entails_unseal => HM HM' ->.
    apply solve_one_not_introducable => //.
  Qed.

End coq_tactics.

Ltac solveOneStep namer_tac :=
  first
  [simple notypeclasses refine (tac_solve_one_fromexist_nested _ _ _ _ _ _ _); [ | | iSolveTC | iSolveTC | simpl]
  |notypeclasses refine (tac_solve_one_fromsep _ _ _ _ _ _ _ _ _ _ _ _ _); [iSolveTC.. | simpl]
  |notypeclasses refine (tac_solve_one_pure _ _ _ _ _ _ _ _ _); [iSolveTC.. | simpl; eauto | iSolveTC]
  |notypeclasses refine (tac_solve_one_not_introducable _ _ _ _ _ _ _ _); [iSolveTC.. | simpl]
  |notypeclasses refine (tac_solve_one_ext_abd_input _ _ _ _ _ _ _ _ _); [iSolveTC | register_delete namer_tac; cbn; simpl] ].





