From iris.proofmode Require Import base coq_tactics reduction tactics.
From diaframe Require Import env_utils util_classes solve_defs tele_utils steps.namer_tacs.
From iris.bi Require Import bi telescopes.

Import bi.

(* This file contains 'case 5' of Diaframe's proof search strategy.
*)

Section solve_sep_rules.

  Context {PROP : bi}.
  Implicit Types M : PROP → PROP.

  Lemma solve_sep_mod_weaker M1 M2 {TT} Q R :
    ModWeaker M1 M2 →
    SolveSep (TT := TT) M1 Q R ⊢ SolveSep (TT := TT) M2 Q R.
  Proof. by unseal_diaframe. Qed.

  Lemma solve_sep_id_emp (R : PROP) :
    R ⊢ SolveSep (TT := [tele]) id emp%I R.
  Proof. unseal_diaframe => /=. eauto. Qed.

  Lemma solve_sep_fromsep {TT : tele} M P R P1 P2 {TT1} P1' {TT2} h M1 M2 :
    (TC∀.. (tt : TT), FromSepCareful (tele_app P tt) (tele_app P1 tt) (tele_app P2 tt)) →
    ExtractDep P1 TT1 P1' TT2 h →
    SplitLeftModality3 M M1 M2 →
    SolveSep M1 P1' $ 
      tele_bind (λ tt1, SolveSep M2 (tele_fun_compose' P2 (tele_app h tt1)) $ tele_fun_compose' R (tele_app h tt1)) 
    ⊢ SolveSep M P R.
  Proof.
    unseal_diaframe; rewrite /TCTForall /FromSepCareful /ExtractDep !tforall_forall => HP HP1 [HM HM1 HM2].
    rewrite -HM => {HM}.
    apply HM1 => {HM1}.
    apply bi_texist_elim => tt1.
    move: HP1 => /(dep_eval tt1) /tforall_forall HP1.
    rewrite tele_app_bind modality_strong_frame_r.
    apply modality_strong_is_mono => {HM2}.
    apply wand_elim_l', bi_texist_elim => tt2.
    apply wand_intro_l.
    rewrite !tele_fun_compose_eq' assoc.
    erewrite HP1, HP.
    by rewrite bi_texist_exist -(exist_intro _).
  Qed.

  Lemma solve_sep_fromexist_nested {TT1 : tele} {TT2} M P P' R :
    (∀.. (tt : TT1), FromTExist (tele_app P tt) (tele_app TT2 tt) (tele_appd_dep P' tt)) →
    ModalityMono M →
    SolveSep (TT := TeleConcatType TT1 TT2) M (tele_curry P') (tele_curry $ tele_dep_bind (λ tt1, tele_bind (λ tt2, tele_app R tt1))) 
    ⊢ SolveSep (TT := TT1) M P R.
  Proof.
    unseal_diaframe => /tforall_forall HPP' HM /=.
    apply HM, bi_texist_elim => tt.
    rewrite (tele_arg_concat_inv_eq tt).
    destruct (tele_arg_concat_inv_strong_hd tt) as [tt1 tt2].
    rewrite bi_texist_exist -(exist_intro tt1).
    rewrite -HPP' !bi_texist_exist.
    rewrite -(exist_intro tt2).
    rewrite tele_uncurry_curry_compose.
    rewrite -tele_curry_uncurry_relation.
    apply sep_mono_r.
    rewrite tele_curry_uncurry_relation.
    rewrite -tele_uncurry_curry_compose.
    rewrite tele_dep_appd_bind.
    by rewrite tele_app_bind.
  Qed.

  Lemma solve_sep_biabd {TTl TTr} M P2 R S T M2 p P1 M1 :
    BiAbd (TTl := TTl) (TTr := TTr) p P1 P2 M2 S T →
    SplitModality3 M M1 M2 →
    □?p P1 ∗ SolveSep (TT := TTl) M1 S $ tele_bind (λ ttl, 
      IntroduceVars (TT := TTr) $ tele_bind (λ ttr, 
        IntroduceHyp (tele_app (tele_app T ttl) ttr) (tele_app R ttr)
    )) ⊢ SolveSep (TT := TTr) M P2 R.
  Proof.
    unseal_diaframe; rewrite /BiAbd => /tforall_forall HP1P2 [HM HM1 HM2].
    rewrite -HM modality_strong_frame_r.
    apply HM1 => {HM1} /=.
    apply wand_elim_l', bi_texist_elim => ttl.
    apply wand_intro_l.
    rewrite tele_app_bind sep_assoc HP1P2 modality_strong_frame_l.
    apply HM2 => {HM2} /=.
    apply wand_elim_l', bi_texist_elim => ttr.
    apply wand_intro_r.
    rewrite (bi_tforall_elim ttr) !tele_app_bind bi_texist_exist -(exist_intro ttr).
    by rewrite -assoc wand_elim_r.
  Qed.

End solve_sep_rules.

Section solve_sep_tacs.
  Context {PROP : bi}.
  Implicit Types M : PROP → PROP.

  Lemma tac_solve_sep_emp Δ M R :
    IntroducableModality M →
    envs_entails Δ R →
    envs_entails Δ $ SolveSep (TT := [tele]) M emp%I R.
  Proof.
    rewrite envs_entails_unseal /= => HM ->.
    by rewrite -solve_sep_mod_weaker -solve_sep_id_emp.
  Qed.

  Lemma tac_solve_sep_fromsep {TT : tele} Δ M M1 M2 P R P1 P2 {TT1} P1' {TT2} h :
    (TC∀.. (tt : TT), FromSepCareful (tele_app P tt) (tele_app P1 tt) (tele_app P2 tt)) →
    ExtractDep P1 TT1 P1' TT2 h →
    SplitLeftModality3 M M1 M2 →
    envs_entails Δ $ SolveSep M1 P1' $ 
      tele_bind (λ tt1, SolveSep M2 (tele_fun_compose' P2 (tele_app h tt1)) $ tele_fun_compose' R (tele_app h tt1)) →
    envs_entails Δ $ SolveSep M P R.
  Proof.
    rewrite envs_entails_unseal => HP1P2 HP1P1' HM ->.
    by eapply solve_sep_fromsep.
  Qed.

  Lemma tac_solve_sep_fromexist_nested {TT1 : tele} Δ {TT2} M P P' R :
    (TC∀.. (tt : TT1), FromTExist (tele_app P tt) (tele_app TT2 tt) (tele_appd_dep P' tt)) →
    ModalityMono M →
    envs_entails Δ $ SolveSep (TT := TeleConcatType TT1 TT2) M (tele_curry P') 
                      (tele_curry $ tele_dep_bind (λ tt1, tele_bind (λ tt2, tele_app R tt1))) →
    envs_entails Δ $ SolveSep (TT := TT1) M P R.
  Proof.
    rewrite envs_entails_unseal => HPP' HM ->.
    by apply solve_sep_fromexist_nested.
  Qed.

  Lemma tac_solve_sep_biabd Δ i M {TTl TTr} P2 R S T M2 p P1 M1 :
    FindInContext Δ (λ p P, BiAbdMod (TTl := TTl) (TTr := TTr) p P P2 M M1 M2 S T) i p P1 →
    envs_entails (envs_delete false i p Δ) 
      (SolveSep (TT := TTl) M1 S (tele_bind $ λ ttl, IntroduceVars (TT := TTr) $ tele_bind (λ ttr, IntroduceHyp (tele_app (tele_app T ttl) ttr) (tele_app R ttr)))) →
    envs_entails Δ $ SolveSep (TT := TTr) M P2 R.
  Proof.
    rewrite envs_entails_unseal => HΔiP HΔ'.
    rewrite (findincontext_spec false) HΔ'.
    apply fic_satisfies in HΔiP as [HPST [HMs HM1 HM2]].
    erewrite solve_sep_biabd => //.
  Qed.

  Lemma tac_solve_sep_ext_biabd Δ mi M {TTl TTr} P2 R S T M2 p P1 M1  :
    FindInExtendedContext Δ (λ p P, BiAbdMod (TTl := TTl) (TTr := TTr) p P P2 M M1 M2 S T) mi p P1 →
    envs_entails (envs_option_delete false mi p Δ)
      (SolveSep (TT := TTl) M1 S (tele_bind $ λ ttl, IntroduceVars (TT := TTr) $ tele_bind (λ ttr, IntroduceHyp (tele_app (tele_app T ttl) ttr) (tele_app R ttr)))) →
    envs_entails Δ $ SolveSep (TT := TTr) M P2 R.
  Proof.
    rewrite envs_entails_unseal => HΔiP HΔ'.
    rewrite (findinextcontext_spec false) // HΔ'.
    apply ficext_satisfies in HΔiP as [HPST [HMs HM1 HM2]].
    erewrite solve_sep_biabd => //.
  Qed.

  Lemma tac_solve_sep_pure Δ {TT : tele} b M P φ Q :
    (TC∀.. tt : TT, FromPure b (tele_app P tt) (tele_app φ tt)) →
    IntroducableModality M →
    (∃.. tt, tele_app φ tt ∧ envs_entails Δ (tele_app Q tt)) →
    envs_entails Δ $ SolveSep (TT := TT) M P Q.
  Proof.
    rewrite envs_entails_unseal texist_exist => /tforall_forall HPφ HM [tt [Hφ ->]].
    unseal_diaframe; rewrite -HM /= bi_texist_exist -(exist_intro tt) -HPφ /= pure_True //=.
    destruct b => /=;
    iIntros "HQ";
    by iSplitR.
  Qed.

End solve_sep_tacs.

Ltac solveSepStep namer_tac := 
  first
  [simple notypeclasses refine (tac_solve_sep_fromexist_nested _ _ _ _ _ _ _ _); [ | | iSolveTC | iSolveTC | simpl]
  |notypeclasses refine (tac_solve_sep_fromsep _ _ _ _ _ _ _ _ _ _ _ _ _ _); 
    [iSolveTC.. | simpl] (* by putting pure after fromsep, we ensure that the pure propositions are not ∧ :)*)
  |notypeclasses refine (tac_solve_sep_pure _ _ _ _ _ _ _ _ _); [iSolveTC.. | simpl]
  |notypeclasses refine (tac_solve_sep_ext_biabd _ _ _ _ _ _ _ _ _ _ _ _ _); 
    [iSolveTC.. | register_delete namer_tac; simpl] ].



















