From iris.proofmode Require Import base coq_tactics reduction tactics.
From diaframe Require Import util_classes util_instances solve_defs tele_utils.
From diaframe.steps Require Import namer_tacs small_steps solve_instances.
From iris.bi Require Import bi telescopes.

Import bi.

(* This file bundles all files in the steps/ folder, to create the iStepS tactic (and friends).
   It also defines notation for Diaframe goals, and adds some support for IPM tactics for these goals.
*)

Export solve_defs.

Lemma from_as_solve_goal {PROP : bi} Δ (P : PROP) P' M P'' :
  TCNoBackTrack (CollectModal P M P') →
  AsSolveGoal M P' P'' →
  envs_entails Δ $ P'' →
  envs_entails Δ P.
Proof. by rewrite envs_entails_unseal /CollectModal => [[<-]] -> ->. Qed.

Lemma as_emp_valid_weak {PROP : bi} (φ : Prop) (P : PROP) `{!AsEmpValidWeak φ P} :
  (⊢ P) → φ.
Proof. by apply as_emp_valid_weak. Qed.

Ltac iStartProof2 :=
  first 
  [ iStartProof 
  | (notypeclasses refine (as_emp_valid_weak _ _ _); [iSolveTC || fail "iStartProof2: not a BI assertion" | iStartProof; simpl]) ].

Ltac iStepS_internal cont :=
  notypeclasses refine (from_as_solve_goal _ _ _ _ _ _ _ _);
    [iSolveTC
    |iSolveTC
    |simpl; cont].

Ltac iStepS :=
  iStartProof2;
  (iStepS_internal ltac:(progress solveSteps)).

Ltac iStepsS :=
  iStartProof2;
  unshelve (progress (repeat (iStepS_internal ltac:(progress solveSteps)))); shelve_unifiable.

Ltac iStepDebug :=
  iStartProof2;
  (iStepS_internal ltac:(idtac)).

Ltac iSmash_int pure_tac :=
  try iStepsS;
  lazymatch goal with
  | |- envs_entails _ _ => ((progress pure_tac); iSmash_int pure_tac) || (iStepDebug; iLeft; iSmash_int pure_tac) || (iStepDebug; iRight; iSmash_int pure_tac)
  | |- _ => try pure_tac
  end.

Tactic Notation "iSmash" "using" tactic(the_tac) := iSmash_int the_tac.

Tactic Notation "iSmash" := iSmash using idtac.


Section iris_tactic_compatibility.
  Context {PROP : bi}.

  (* Makes iExist work on SolveSep goals *)
  Global Instance solve_sep_into_exist `{TTf : A → tele} M (P Q : TeleS TTf -t> PROP) :
    ModalityStrongMono M →
    FromExist (SolveSep (TT := TeleS TTf) M P Q) (λ a, SolveSep (TT := TTf a) M (P a) (Q a)).
  Proof.
    unseal_diaframe; rewrite /FromExist => HM.
    apply exist_elim => a.
    apply HM.
    rewrite !bi_texist_exist.
    apply exist_elim => tt.
    by rewrite -(exist_intro $ TargS a tt) /=.
  Qed.

  (* Makes iIntros "H" work on IntroduceHyp *)
  Global Instance introduce_hyp_into_wand (Q R : PROP) :
    FromWand (IntroduceHyp Q R) Q R.
  Proof. unseal_diaframe. iSolveTC. Qed.

  (* Makes iIntros (x) work on IntroduceVars *)
  Global Instance introduce_vars_into_forall {A : Type} {b : A → tele} (Q : TeleS b -t> PROP) :
    FromForall (IntroduceVars Q) (λ a : A, IntroduceVars (TT := b a) (Q a)) (λ x, x).
  Proof. unseal_diaframe; iSolveTC. Qed.

  (* Makes iMod work on SolveSep *)
  Global Instance solve_sep_elim_modal p (P : PROP) P' M M1 M2 {TT} R S T :
    IntoModal p P M1 P' →
    TCEq T (SolveSep M (TT := TT) R S) →
    SplitModality3 M M1 M2 →
    ElimModal True p false P P' T (SolveSep M2 (TT := TT) R S).
  Proof.
    unseal_diaframe; rewrite /IntoModal /ElimModal /= => -> -> [<- HM1 HM2] _.
    rewrite modality_strong_frame_l.
    apply HM1.
    by rewrite wand_elim_r.
  Qed.

  (* Makes iSplitL/iSplitR work on SolveSep *)
  Global Instance solvesep_fromsep (L G : PROP) M :
    ModalityStrongMono M →
    FromSep (SolveSep (TT := TeleO) M L G) (M L) G | 999. 
    (* note that G will have its own modality by construction *)
  Proof.
    rewrite /FromSep. iStepsS. unseal_diaframe => /=.
    rewrite -modality_strong_frame_l. eauto with iFrame.
  Qed.

  (* Makes iLeft/iRight work on SolveSep *)
  Lemma solve_sep_first_from_or {TT : tele} (M : PROP → PROP) P P1 P2 R :
    (TC∀.. tt : TT, FromOr (tele_app P tt) (tele_app P1 tt) (tele_app P2 tt)) →
    ModalityMono M →
    FromOr (SolveSep (TT := TT) M P R) (SolveSep (TT := TT) M P1 R) (SolveSep (TT := TT) M P2 R).
  Proof.
    unseal_diaframe; rewrite /FromOr => /tforall_forall HPs HM.
    apply bi.or_elim; apply HM; apply bi_texist_mono => tt.
    all: apply sep_mono_l; rewrite -HPs.
    - rewrite -bi.or_intro_l //.
    - rewrite -bi.or_intro_r //.
  Qed.

  (* Makes iLeft/iRight work on SolveOne *)
  Lemma solve_one_from_or {TT : tele} (M : PROP → PROP) P P1 P2 :
    (TC∀.. tt : TT, FromOr (tele_app P tt) (tele_app P1 tt) (tele_app P2 tt)) →
    ModalityMono M →
    FromOr (SolveOne (TT := TT) M P) (SolveOne (TT := TT) M P1) (SolveOne (TT := TT) M P2).
  Proof.
    unseal_diaframe; rewrite /FromOr => /tforall_forall HPs HM.
    apply bi.or_elim; apply HM; apply bi_texist_mono => tt.
    all: rewrite -HPs.
    - rewrite -bi.or_intro_l //.
    - rewrite -bi.or_intro_r //.
  Qed.

  (* Makes iInv work on SolveSep *)
  Global Instance solvesep_elimacc `{!BiFUpd PROP} E E1 E2 {X : Type} α β mγ {TT} (R : TT -t> PROP) Q :
    ElimAcc (X := X) True (fupd E1 E2) (fupd E2 E1) α β mγ (SolveSep (TT := TT) (fupd E1 E) R Q) 
      (λ x, SolveSep (TT := [tele]) (fupd E2 E2) (β x) (IntroduceHyp (default emp (mγ x))%I (SolveSep (TT := TT) (fupd E1 E) R Q))).
  Proof.
    unseal_diaframe; rewrite /ElimAcc /accessor => _ /=.
    iIntros "Hx1 Hx2".
    iMod "Hx2" as (x) "[Hα Hβ]".
    iSpecialize ("Hx1" with "Hα").
    iMod "Hx1" as "[Hβ' Hr]".
    iSpecialize ("Hβ" with "Hβ'").
    iMod "Hβ".
    iSpecialize ("Hr" with "Hβ") => //.
  Qed.

  (* Makes iMod work on SolveSep (TODO: wait what does [solve_sep_elim_modal] then accomplish?) *) 
  Global Instance solve_sep_elim_modal_fupd `{!BiFUpd PROP}  (P : PROP) E1 E2 E3 p {TT} A G: 
    ElimModal True p false (|={E1, E2}=> P)%I P (SolveSep (TT := TT) (fupd E1 E3) A G) (SolveSep (TT := TT) (fupd E2 E3) A G).
  Proof.
    rewrite /ElimModal /= bi.intuitionistically_if_elim => _; iStepS.
    unseal_diaframe. iMod "H1".
    by iMod ("H2" with "H1") as "$".
  Qed.

  (* Makes iMod work on SolveSep *) 
  Global Instance solve_sep_elim_modal_bupd `{BiBUpdFUpd PROP} (P : PROP) E1 E2 p {TT} A G: 
    ElimModal True p false (|==> P)%I P (SolveSep (TT := TT) (fupd E1 E2) A G) (SolveSep (TT := TT) (fupd E1 E2) A G).
  Proof.
    rewrite /ElimModal /= bi.intuitionistically_if_elim => _; iStepS.
    unseal_diaframe. iMod "H1".
    by iMod ("H2" with "H1").
  Qed.

  (* Used for the iReIntro tactic, which reruns a hypothesis through Diaframe's introduction steps *)
  Lemma prepend_emp {TT} (A : TT -t> PROP) G M M1 M2 Δ:
    SplitLeftModality3 M M1 M2 →
    envs_entails Δ $ SolveSep (TT := [tele]) M1 emp%I (emp ∗ SolveSep (TT := TT) M2 A G)%I →
    envs_entails Δ $ SolveSep M A G.
  Proof.
    unseal_diaframe; rewrite envs_entails_unseal.
    case => [HM HM1 HM2] -> /=.
    rewrite -HM.
    apply HM1. by rewrite !left_id.
  Qed.

End iris_tactic_compatibility.

Global Ltac iReIntro pat_name :=
  try (eapply prepend_emp; [iSolveTC | ]);
  iRevert pat_name;
  iStepS;
  try (rewrite left_id). (* breaks in some cases but usually sufficient *)

Global Hint Extern 4 (FromOr (SolveSep _ _ _) _ _) =>
    eapply solve_sep_first_from_or; iSolveTC : typeclass_instances.

Global Hint Extern 4 (FromOr (SolveOne _ _) _ _) =>
    eapply solve_one_from_or; iSolveTC : typeclass_instances.

Notation "'‖' M '‖' ∃ x1 .. xn '‖' L '(∗)' R" :=
  (SolveSep 
    (TT := TeleS (λ x1, .. (TeleS (λ xn, TeleO)) ..))
    M
    (λ x1, .. (λ xn, L%I) .. )
    (λ x1, .. (λ xn, R%I) .. )
  )
  (at level 20, M at level 9, x1 binder, xn binder, L, R at level 200, format
  "‖  M  ‖  ∃  x1 .. xn  ‖  L  (∗)  R"
).


Notation "'‖' M '‖' '∃' '-' '‖' L '(∗)' R" :=
  (SolveSep 
    (TT := TeleO)
    M
    L
    R
  )
  (at level 20, M at level 9, L, R at level 200, format
  "‖  M  ‖  ∃  -  ‖  L  (∗)  R"
).

Notation "'‖' M '‖' ∃ x1 .. xn '‖' L" :=
  (SolveOne 
    (TT := TeleS (λ x1, .. (TeleS (λ xn, TeleO)) ..))
    M
    (λ x1, .. (λ xn, L%I) .. )
  )
  (at level 20, M at level 9, x1 binder, xn binder, L at level 200, format
  "‖  M  ‖  ∃  x1 .. xn  ‖  L"
).

Notation "'‖' M '‖' '∃' '-' '‖' L" :=
  (SolveOne 
    (TT := TeleO)
    M
    L
  )
  (at level 20, M at level 9, L at level 200, format
  "‖  M  ‖  ∃  -  ‖  L"
).

Notation "'‖' ∀ x1 .. xn ',' G" :=
  (IntroduceVars
    (TT := TeleS (λ x1, .. (TeleS (λ xn, TeleO)) ..))
    (λ x1, .. (λ xn, G%I) .. )
  )
  (at level 20, x1 binder, xn binder, G at level 200, format
  "‖  ∀  x1 .. xn  ,  G"
).

Notation "'‖' ∀ '-' ',' G" :=
  (IntroduceVars
    (TT := TeleO)
    G
  )
  (at level 20, G at level 200, format
  "‖  ∀  -  ,  G"
).

Notation "'‖' L (-∗) R" :=
  (IntroduceHyp L R)
  (at level 20, L, R at level 200, format
  "‖  L  (-∗)  R"
).

















