From diaframe.heap_lang Require Import proof_automation.

From iris.algebra Require Import lib.frac_auth numbers. 
From iris.heap_lang Require Import proofmode.
From iris.heap_lang Require Import lib.par lib.counter lib.lock.

Import bi.

From diaframe.lib Require Import own_hints.


Section lock_increment.
  Context `{!heapGS Σ, !spawnG Σ, !ccounterG Σ}.
  Context (L : lock).
  Local Existing Instance ccounter_inG.

  Definition parallel_add : val := λ: "s",
    let: "l" := ref "s" in
    let: "lk" := newlock L #() in
    ((acquire L "lk";; "l" <- #1 + ! "l" ;; release L "lk") ||| 
     (acquire L "lk";; "l" <- #1 + ! "l" ;; release L "lk")) ;;
    acquire L "lk" ;;
    ! "l".

  Global Opaque par.

  Proposition parallel_add_spec (s : nat) :
    {{{ True }}} parallel_add #s {{{ RET #(2 + s); True }}}.
  Proof.
    iStepsS.
    iAssert (|==> ∃ γ, (∃ (m : nat), x0 ↦ #(m + s) ∗ own γ (●F m)) ∗ own γ (◯F 0))%I with "[H2]" 
      as ">[%γ [HR Hγ']]"; first iStepsS. iIntros "!>".
    wp_apply (newlock_spec _ (∃ m : nat, x0 ↦ #(m + s) ∗ own γ (●F m)) with "HR").
    iStepsS.
    iDestruct "Hγ'" as "[Hγ1 Hγ2]". iIntros "!>".
    wp_apply (wp_par (λ _, own γ (◯F{1/2} 1)) (λ _, own γ (◯F{1/2} 1)) with "[Hγ1] [Hγ2]");
    iStepsS.
  Qed.

  Definition parallel_add_fg : val := λ: "s",
    let: "l" := ref "s" in
    ((FAA "l" #1) ||| 
     (FAA "l" #1)) ;;
    ! "l".

  Proposition parallel_add_fg_spec (s : nat) :
    {{{ True }}} parallel_add_fg #s {{{ RET #(2%nat + s); True }}}.
  Proof.
    do 18 iStepS. (* par will see that it knows a spec for FAA, and shortcircuit. We need to prevent that *)
    pose proof (nroot .@ "N") as N.
    iAssert (|={⊤}=> ∃ γ, inv N (∃ m : nat, x0 ↦ #(m + s) ∗ own γ (●F m)) ∗ own γ (◯F 0))%I with "[H2]"
      as ">[%γ [#HN Hγ']]"; first iStepsS.
    iDestruct "Hγ'" as "[Hγ1 Hγ2]". iIntros "!>".
    wp_apply (wp_par (λ _, own γ (◯F{1/2} 1)) (λ _, own γ (◯F{1/2} 1)) with "[Hγ1] [Hγ2]");
    iStepsS.
  Qed.

  Definition parallel_store : val := λ: <>,
    let: "l" := ref #5 in
    (("l" <- #3) ||| ("l" <- #7));;
    ! "l".

  Proposition parallel_store_spec : {{{ True }}} parallel_store #() {{{ (n : Z), RET #n; True }}}.
  Proof.
    do 18 iStepS.
    pose proof (nroot .@ "N") as N.
    iAssert (|={⊤}=> inv N (∃ n : Z, x0 ↦ #n))%I with "[H2]" as ">#HN"; first iStepsS.
    iIntros "!>". wp_apply (wp_par (λ _, True) (λ _, True))%I;
    iStepsS.
  Qed.

  Proposition free_inv_loop P : {{{ P }}} let: "l" := ref #2 in Free "l" {{{ RET #(); P }}}.
  Proof.
    do 8 iStepS.
    iMod (inv_alloc nroot _ (x0 ↦ #2 ∗ P) with "[H3 H1]") as "#HN". { iStepsS. }
    iStepsS. (* this tests that this terminates *)
  Abort.

End lock_increment.


From iris.algebra Require Import csum agree excl.

Definition stateR := csumR (exclR unitO) (agreeR unitO).
Class stateG Σ := {state_inG :> inG Σ stateR}.
Definition stateΣ : gFunctors := #[GFunctor stateR].
Local Instance sub_stateΣ Σ : subG stateΣ Σ → stateG Σ.
Proof. solve_inG. Qed.

Definition Start : stateR := Cinl (Excl ()).
Definition Finish : stateR := Cinr (to_agree ()).

Class fracG Σ := { frac_inG :> inG Σ fracR }.
Definition fracΣ : gFunctors := #[GFunctor fracR].
Local Instance subG_fracΣ {Σ} : subG fracΣ Σ → fracG Σ.
Proof. solve_inG. Qed.


Section stateG_automation.
  Context `{!stateG Σ}.

  Global Instance biabd_start_finish p γ :
    HINT □⟨p⟩ own γ Start ✱ [- ; emp] ⊫ [bupd]; own γ Finish ✱ [own γ Finish].
  Proof. 
    iStepS. rewrite bi.intuitionistically_if_elim.
    rewrite own_update; last by apply (cmra_update_exclusive Finish).
    (* TODO: iStepS does not solve this, due to the different sum_inr_subtract_n/s instances *)
    iMod "H1" as "#H". 
    iStepsS.
  Qed.

  Global Instance finish_dup_inst γ : Persistent (own γ Finish).
  Proof.
    apply own_core_persistent.
    rewrite /CoreId. done.
  Qed.

  Global Instance state_valid_op_both_finish :
    IsValidOp Finish Finish Finish Σ True%I.
  Proof. split; eauto. iStepsS. rewrite /Finish -Cinr_op agree_idemp //. Qed.
  Global Instance state_valid_op_left_start (s : stateR) :
    IsValidOp (Start ⋅ s) Start s Σ False%I.
  Proof. split; destruct s; eauto. Qed.
  Global Instance state_valid_op_right_start (s : stateR) :
    IsValidOp (s ⋅ Start) s Start Σ False%I.
  Proof. split; destruct s; eauto. Qed.
End stateG_automation.


Definition slow_incr : val := 
  λ: "l",
    "l" <- ! "l" + # 1.

Definition slow_incr_par : val :=
  λ: "l",
    (slow_incr "l" ||| slow_incr "l");;
    ! "l".


Section exercises.
  Context `{!heapGS Σ, !stateG Σ, !fracG Σ, !spawnG Σ}.

  Let N := nroot.

  Definition double_incr_inv (γ1 γ2 : gname) (l : loc) (n : Z) : iProp Σ :=
    ∃ m : Z, l ↦ # m ∗ ((⌜m = n⌝ ∗ own γ1 Start)
                        ∨ (⌜m = (n + 1)%Z⌝ ∗ own γ1 Finish ∗ own γ2 (1/2)%Qp)
                        ∨ (⌜m = (n + 2)%Z⌝ ∗ own γ1 Finish ∗ own γ2 1%Qp)).

  Instance slow_incr_spec (n : Z) (l : loc) :
    SPEC γ1 γ2, {{ inv N (double_incr_inv γ1 γ2 l n) ∗ own γ2 (1/2)%Qp }}
      slow_incr #l
    {{ RET #(); own γ1 Finish }}.
  Proof. iStepsS. Qed.

  Instance slow_incr_par_spec (n : Z) (l : loc) :
    SPEC {{ l ↦ #n }}
      slow_incr_par #l
    {{ (z : Z), RET #z; ⌜z = (n + 1)%Z⌝ ∨ ⌜z = (n + 2)%Z⌝ }}.
  Proof.
    do 8 iStepS.
    (* allocation is manual since γ2 is not mentioned in the first case of invariant, and we have no CoAllocate instance.
       could be fixed with a (1/3) mention. but alright *)
    iMod (own_alloc Start) as (γ1) "Hγ1"; first done.
    iMod (own_alloc 1%Qp) as (γ2) "Hγ2"; first done.
    iSmash.
  Qed.
End exercises.


































