From diaframe.heap_lang Require Import proof_automation atomic_specs wp_auto_lob.
From iris.heap_lang Require Import proofmode.
From iris.algebra Require Import frac_auth numbers excl.
From diaframe.lib Require Import ticket_cmra frac_token own_hints.

Definition make_barrier : val := 
  λ: <>, ref #0.

Definition sync_up_enter : val := rec: "sync_enter" "b" :=
  let: "z" := ! "b" in
  if: #0 ≤ "z" then
    if: CAS "b" "z" ("z" + #1) then 
      "z" 
    else 
      "sync_enter" "b"
  else "sync_enter" "b".

Definition wait_for : val :=
  rec: "wait_for" "b" "w" := 
    let: "z" := ! "b" in
    if: "z" = "w" then
      #()
    else
      "wait_for" "b" "w".

Definition sync_up_exit threads : val := 
  λ: "b" "v",
    let: "w" := 
      if: "v" = #0 then 
        #(Zpos threads) 
      else 
        "v" 
      in
    wait_for "b" "w";;
    "b" <- "w" - #1.

Definition sync_up threads : val := λ: "b",
  sync_up_exit threads "b" (sync_up_enter "b").

Definition sync_down_enter : val := rec: "sync_enter" "b" :=
  let: "z" := ! "b" in
  if: "z" ≤ #0 then
    if: CAS "b" "z" ("z" - #1) then
     - "z" 
    else 
      "sync_enter" "b"
  else 
    "sync_enter" "b".

Definition sync_down_exit threads : val := 
  λ: "b" "v",
    let: "w" := 
      if: "v" = #0 then 
        #(Zneg threads) 
      else 
        - "v" 
      in
    wait_for "b" "w";;
    "b" <- "w" + #1.

Definition sync_down threads : val := λ: "b",
  sync_down_exit threads "b" (sync_down_enter "b").

Definition barrierUR := authUR $ optionUR $ exclR $ leibnizO Z.
Class barrierG Σ :=
  BarrierG { 
    barrier_ticketG :> inG Σ ticketUR;
    barrier_tokenG :> fracTokenG Σ;
    barrier_barrierG :> inG Σ barrierUR;
  }.
Definition barrierΣ : gFunctors :=
  #[GFunctor ticketUR; fracTokenΣ; GFunctor barrierUR].

Local Obligation Tactic := smash_verify_tac.

Global Program Instance subG_barrierΣ {Σ} : subG barrierΣ Σ → barrierG Σ.

Section proof.
  Context `{!heapGS Σ}.
  Set Default Proof Using "heapGS0".

  Instance sync_up_enter_phys_spec (l : loc) :
    SPEC (z : Z), << ▷ l ↦ #z >>
      sync_up_enter #l
    << RET #z; l ↦ #(z + 1) ∗ ⌜0 ≤ z⌝%Z>>.
  Proof. iSmash. Qed.

  Instance sync_down_enter_phys_spec (l : loc) :
    SPEC (z : Z), << ▷ l ↦ #z >>
      sync_down_enter #l
    << RET #(-z); l ↦ #(z - 1) ∗ ⌜z ≤ 0⌝%Z>>.
  Proof. iSmash. Qed.

  Instance sync_up_wait_for_spec (l : loc) (zw : Z) :
    SPEC (dq : dfrac) (z : Z), << ▷ l ↦{dq} #z >>
      wait_for #l #zw
    << RET #(); l ↦{dq} #z ∗ ⌜z = zw⌝ >>.
  Proof.
    iStepsS. destruct (decide (x6 = x2)) as [-> | Hneq]; iSmash.
  Qed.

  Context (threads : positive).
  Context `{!barrierG Σ} (P1 P2 : Qp → iProp Σ).
  Context `{!Fractional P1, !Fractional P2}.
  Let N := nroot .@ "barrier".
  Set Default Proof Using "threads heapGS0 barrierG0 P1 P2 Fractional0 Fractional1".

  Definition barrier_inv γp1 γp2 γt γb (l : loc) : iProp Σ :=
    ∃ (z : Z), l ↦ #z ∗ own γt (CoPset $ tickets_geq (S (Z.abs_nat z))) ∗ own γb (● Excl' z) ∗ (
      (⌜0 ≤ z⌝%Z ∗ (
          (⌜z < Zpos threads⌝%Z ∗ token_counter P1 γp1 (Z.to_pos (Zpos threads - z)) ∗ no_tokens P2 γp2 1%Qp ∗ own γt (CoPset $ ticket (Z.abs_nat z)) ∗ own γb (◯ Excl' z))
                ∨ 
          (⌜z < Zpos threads⌝%Z ∗ no_tokens P1 γp1 1%Qp ∗ token_counter P2 γp2 (Z.to_pos (Zpos threads - z)) ∗ (own γb (◯ Excl' z) ∨ own γt (CoPset $ ticket (Z.abs_nat z))) ∗ own γt (CoPset $ ticket 0))
                ∨
          (⌜z = Zpos threads⌝ ∗ no_tokens P1 γp1 1%Qp ∗ no_tokens P2 γp2 1%Qp ∗ P1 1 ∗ (own γb (◯ Excl' z) ∨ own γt (CoPset $ ticket 0)) ∗ own γt (CoPset $ ticket $ Pos.to_nat threads))))
        ∨ 
      (⌜z ≤ 0⌝%Z ∗ (
          (⌜Zneg threads < z⌝%Z ∗ token_counter P2 γp2 (Z.to_pos (Zpos threads + z)) ∗ no_tokens P1 γp1 1%Qp ∗ own γt (CoPset $ ticket (Z.abs_nat z)) ∗ own γb (◯ Excl' z))
                ∨ 
          (⌜Zneg threads < z⌝%Z ∗ no_tokens P2 γp2 1%Qp ∗ token_counter P1 γp1 (Z.to_pos (Zpos threads + z)) ∗ (own γb (◯ Excl' z) ∨ own γt (CoPset $ ticket (Z.abs_nat z))) ∗ own γt (CoPset $ ticket 0))
                ∨
          (⌜z = Zneg threads⌝ ∗ no_tokens P1 γp1 1%Qp ∗ no_tokens P2 γp2 1%Qp ∗ P1 1 ∗ (own γb (◯ Excl' z) ∨ own γt (CoPset $ ticket 0)) ∗ own γt (CoPset $ ticket (Pos.to_nat threads)))))).

  Definition is_barrier γp1 γp2 γt γb (v : val) : iProp Σ :=
    ∃ (l : loc), ⌜v = #l⌝ ∗ □ (P1 1 ={⊤∖↑N}=∗ P2 1) ∗ □ (P2 1 ={⊤∖↑N}=∗ P1 1) ∗ inv N (barrier_inv γp1 γp2 γt γb l).

  Program Instance p1p2fupd_atom : AtomAndConnective true (P1 1 ={⊤∖↑N}=∗ P2 1).
  Program Instance p2p1fupd_atom : AtomAndConnective true (P2 1 ={⊤∖↑N}=∗ P1 1).

  Ltac connective_as_atom_shortcut ::= biabd_step_atom_shortcut.

  Instance newbarrier_spec  :
    SPEC {{ P1 1 ∗ □ (P1 1 ={⊤∖↑N}=∗ P2 1) ∗ □ (P2 1 ={⊤∖↑N}=∗ P1 1) }} 
      make_barrier #()
    {{ γp1 γp2 γt γb (v : val), RET v; is_barrier γp1 γp2 γt γb v ∗ token_iter P1 (Pos.to_nat threads) γp1 }}.
  Proof.
    iStepsS.
    iLeft; iLeft.
    erewrite (right_id ∅); last iSolveTC.
    iStepsS.
  Qed.

  (* this is ugly and currently required since atomic specs above do not expose the facts
     that the linearization point happens at a physical step, where we can strip a later.
     TODO: either strengthen some atomic specs to expose this, or wait for later credits *)
  Instance token_counter_timeless `{!Fractional P} γ p : Timeless (token_counter P γ p).
  Proof. Admitted.
  Instance token_timeless `{!Fractional P} γ : Timeless (token P γ).
  Proof. Admitted.

  Instance sync_up_enter_spec γp1 γp2 γt γb (v : val) :
    SPEC {{ is_barrier γp1 γp2 γt γb v ∗ token P1 γp1 }}
      sync_up_enter v
    {{ (z : Z), RET #z; ⌜0 ≤ z⌝%Z ∗ own γt (CoPset $ ticket (Z.to_nat z)) }}.
  Proof.
    iStepsS; try iSmash.
    - destruct (decide (x1 + 1 = Z.pos threads)%Z) as [Heq|Hneq]; iSmash.
    - destruct (decide (threads = xH)) as [Heq|Hneq]; iSmash.
  Qed.

  Instance sync_up_exit_spec  γp1 γp2 γb γt (z : Z) (v : val) :
    SPEC {{ is_barrier γp1 γp2 γt γb v ∗ ⌜0 ≤ z⌝%Z ∗ own γt (CoPset $ ticket $ Z.to_nat z) }}
      sync_up_exit threads v #z
    {{ RET #(); token P2 γp2 }}.
  Proof. iSmash. Qed. 

  Global Instance sync_up_spec γp1 γp2 γt γb (v : val) :
    SPEC {{ is_barrier γp1 γp2 γt γb v ∗ token P1 γp1 }}
      sync_up threads v
    {{ RET #(); token P2 γp2 }}.
  Proof.
    Local Existing Instance defs.as_template_step | 0.
    iStepsS.
  Qed.

  Instance sync_down_enter_spec γp1 γp2 γt γb (v : val) :
    SPEC {{ is_barrier γp1 γp2 γt γb v ∗ token P2 γp2 }}
      sync_down_enter v
    {{ (z : Z), RET #z; ⌜0 ≤ z⌝%Z ∗ own γt (CoPset $ ticket (Z.to_nat z)) }}.
  Proof.
    iStepsS; try iSmash. (* iSmash deals with failing CAS's *)
    - replace (⊤ ∖ ∅ ∖ ↑N) with (⊤ ∖ ↑N : coPset); last set_solver.
      destruct (decide (x1 - 1 = Z.neg threads)%Z) as [Heq|Hneq]; iSmash.
    - replace (⊤ ∖ ∅ ∖ ↑N) with (⊤ ∖ ↑N : coPset); last set_solver.
      destruct (decide (x1 - 1 = Z.neg threads)%Z) as [Heq|Hneq]; iSmash.
  Qed.

  Instance sync_down_exit_spec γp1 γp2 γb γt (z : Z) (v : val) :
    SPEC {{ is_barrier γp1 γp2 γt γb v ∗ ⌜0 ≤ z⌝%Z ∗ own γt (CoPset $ ticket $ Z.to_nat z) }}
      sync_down_exit threads v #z
    {{ RET #(); token P1 γp1 }}.
  Proof. iSmash. Qed.

  Global Instance sync_down_spec γp1 γp2 γt γb (v : val) :
    SPEC {{ is_barrier γp1 γp2 γt γb v ∗ token P2 γp2 }}
      sync_down threads v
    {{ RET #(); token P1 γp1 }}.
  Proof.
    Local Existing Instance defs.as_template_step | 0.
    iStepsS.
  Qed.
End proof.


Section barrier_specs2. (* this is currently necessary since otherwise TC search will try to find fractionals :( *)
  Context `{!heapGS Σ, barrierG Σ}. 
  Context (P1 P2 : Qp → iProp Σ).

  Global Program Instance barrier_persistent_reuse threads γp1 γp2 γt γb (v : val) `{!Fractional P1, !Fractional P2} :
    Persistent (is_barrier threads P1 P2 γp1 γp2 γt γb v).

  Global Opaque is_barrier make_barrier sync_up sync_down.

  Global Instance sync_up_spec_frac_reuse threads γp1 γp2 γt γb (v : val) :
    SPEC HP1 HP2, {{ is_barrier threads P1 P2 (Fractional0 := HP1) (Fractional1 := HP2) γp1 γp2 γt γb v ∗ token P1 γp1 }}
      sync_up threads v
    {{ RET #(); token P2 γp2 }}.
  Fail Proof. (* TODO: what is going on here..? *)
    do 3 iStepS. assert (H1 := sync_up_spec threads P1 P2). iStepsS. 
  Qed.

  Global Instance sync_down_spec_frac_reuse threads γp1 γp2 γt γb (v : val) :
    SPEC HP1 HP2, {{ is_barrier threads P1 P2 (Fractional0 := HP1) (Fractional1 := HP2) γp1 γp2 γt γb v ∗ token P2 γp2 }}
      sync_down threads v
    {{ RET #(); token P1 γp1 }}.
  Fail Proof. 
    do 3 iStepS. assert (H1 := sync_down_spec threads P1 P2). iStepsS. 
  Qed.
End barrier_specs2.

