From diaframe.heap_lang Require Import proof_automation wp_auto_lob.
From diaframe.lib Require Import frac_token.

Definition new_rwlock : val :=
  λ: "_", ref (#0).

Definition acquire_reader : val :=
  rec: "acquire_rdr" "rwlk" :=
    let: "z" := ! "rwlk" in
    if: "z" = #-1 then
      "acquire_rdr" "rwlk"
    else 
      if: CAS "rwlk" "z" ("z" + #1) then 
        #() 
      else 
        "acquire_rdr" "rwlk".

Definition release_reader : val :=
  λ: "rwlk", (FAA "rwlk" #-1) ;; #().

Definition acquire_writer : val :=
  rec: "acquire_wrtr" "rwlk" :=
    if: CAS "rwlk" (#0) #-1 then 
      #() 
    else 
      "acquire_wrtr" "rwlk".

Definition release_writer : val :=
  λ: "rwlk", "rwlk" <- #0.

Section proof.
  Context `{!heapGS Σ, fracTokenG Σ}.
  Context (P : Qp → iProp Σ) {HP : Fractional P}.
  Let N := nroot .@ "rw_lock".

  Instance nogenP : do_lob.NoLobGen P := I.
  Instance nogenHP : do_lob.NoLobGen HP := I.

  Obligation Tactic := program_verify.

  Definition rwlock_inv γt (rw_loc : loc) : iProp Σ :=
    ∃ (z : Z), 
      rw_loc ↦ #z ∗
        ((⌜0 < z⌝%Z ∗ token_counter P γt (Z.to_pos z)) ∨
         (⌜z = 0⌝%Z ∗ no_tokens P γt 1 ∗ P 1) ∨
         (⌜z = -1⌝%Z ∗ no_tokens P γt (1/2)%Qp)).

  Definition is_rwlock γt rwlk : iProp Σ :=
    ∃ (rw_loc : loc), 
      ⌜rwlk = #rw_loc⌝ ∗ inv N (rwlock_inv γt rw_loc).

  Program Instance new_rwlock_spec :
    SPEC {{ P 1 }} 
      new_rwlock #() 
    {{ (rwlk : val) γt, RET rwlk; is_rwlock γt rwlk }}.

  Global Program Instance acquire_reader_spec γt (rwlk : val) :
    SPEC {{ is_rwlock γt rwlk }} 
      acquire_reader rwlk 
    {{ RET #(); token P γt }}.

  Global Instance release_reader_spec γt (rwlk : val) :
    SPEC {{ is_rwlock γt rwlk ∗ token P γt }}
      release_reader rwlk
    {{ RET #(); True }}.
  Proof. iStepsS. destruct (decide (x1 = 1)); iStepsS. Qed.

  Global Program Instance acquire_writer_spec γt (rwlk : val) :
    SPEC {{ is_rwlock γt rwlk }}
      acquire_writer rwlk
    {{ RET #(); P 1 ∗ no_tokens P γt (1/2)%Qp }}.

  Global Program Instance release_writer_spec γt (rwlk : val) :
    SPEC {{ is_rwlock γt rwlk ∗ P 1 ∗ no_tokens P γt (1/2)%Qp }}
      release_writer rwlk 
    {{ RET #(); True }}.
End proof.



















