From iris.algebra Require Import excl.
From diaframe.heap_lang Require Import proof_automation atomic_specs wp_auto_lob.
From diaframe.lib Require Import own_hints.
From iris.heap_lang Require Import proofmode notation atomic_heap.
Set Default Proof Using "Type".

Import bi notation.

Class lockG Σ := LockG { lock_tokG :> inG Σ (exclR unitO) }.
Definition lockΣ : gFunctors := #[GFunctor (exclR unitO)].

Local Obligation Tactic := program_smash_verify.

Global Program Instance subG_lockΣ {Σ} : subG lockΣ Σ → lockG Σ.

Section proof.
  Context `{!heapGS Σ, !lockG Σ} {aheap: atomic_heap}.

  Definition newlock : val := 
    λ: <>, ref #false.

  Definition acquire : val :=
    rec: "acquire" "l" := 
      if: CAS "l" #false #true then 
        #() 
      else 
        "acquire" "l".

  Definition release : val := 
    λ: "l", "l" <- #false.

  Definition lock_state (γ : gname) (l : loc) (b : bool) : iProp Σ :=
    l ↦ #b ∗ (⌜b = true⌝ ∨ ⌜b = false⌝ ∗ own γ (Excl ())).

  Definition locked (γ : gname) : iProp Σ :=
    own γ (Excl ()).

  Global Program Instance newlock_spec :
    SPEC {{ True }}
      newlock #()
    {{ (l : loc) γ, RET #l; lock_state γ l false}}.

  Global Program Instance acquire_spec (l : loc) γ :
    SPEC b, << ▷ lock_state γ l b >>
        acquire #l
    << RET #(); lock_state γ l true ∗ locked γ ∗ ⌜b = false⌝ >>.

  Global Program Instance release_spec (l : loc) γ :
    SPEC b, << ▷ lock_state γ l b ∗ ▷ locked γ >>
        release #l
    << RET #(); lock_state γ l false ∗ ⌜b = true⌝ >>.

  Global Program Instance locked_timeless γ : Timeless (locked γ).
End proof.

Global Opaque lock_state locked acquire release newlock.
