From iris.algebra Require Import coPset frac_auth agree.
From diaframe.heap_lang Require Import proof_automation atomic_specs wp_auto_lob.
From iris.heap_lang Require Import proofmode notation atomic_heap.
From diaframe.lib Require Import ticket_cmra own_hints.
From diaframe.heap_lang.examples.logatom Require Import atomic_cas_counter.

Definition tlockUR := frac_authUR $ optionUR $ agreeR $ natO.
Class tlockG Σ := {
  tlock_ticketG :> inG Σ ticketUR ;
  tlock_tlockG :> inG Σ tlockUR;
}.
Definition tlockΣ : gFunctors :=
  #[ GFunctor ticketUR; GFunctor tlockUR ].

Local Obligation Tactic := program_smash_verify.

Global Program Instance subG_tlockΣ {Σ} : subG tlockΣ Σ → tlockG Σ.

Import bi notation.

Section proof.
  Context `{!heapGS Σ, !tlockG Σ} {aheap: atomic_heap}.

  Definition wait_equal : val :=
    rec: "wait_loop" "x" "l" :=
      let: "o" := ! "l" in
      if: "x" = "o" then 
        #() (* my turn *)
      else 
        "wait_loop" "x" "l".

  Definition newlock : val :=
    λ: <>, ((* owner *) ref #0, (* next *) ref #0).

  Definition acquire : val :=
    λ: "lk",
      let: "ticket" := incr (Snd "lk") in
      wait_equal "ticket" (Fst "lk").

  Definition release : val :=
    λ: "lk", weak_incr (Fst "lk");; #().

  Instance wait_equal_spec (l : loc) (z : Z) :
    SPEC w, << ▷ l ↦ w >>
      wait_equal #z #l
    << RET #(); l ↦ #z ∗ ⌜w = #z⌝ >>.
  Proof.
    iStepsS; last iSmash.
    destruct (decide (x5 = #x3)); iSmash.
  Qed.

  Definition ticket_inv (l : loc) γ : iProp Σ := ∃ (n : nat), l ↦ #n ∗ own γ (CoPset $ tickets_geq n).
  Definition issued (γ : gname) (x : nat) : iProp Σ := own γ (CoPset $ ticket x).
  Definition locked_at γ1 (n : nat) : iProp Σ := own γ1 (◯F (Some $ to_agree n)).
  Definition lock_state γ1 γ2 l1 l2 (n : nat) (b : bool) : iProp Σ :=
    l1 ↦ #n ∗ own γ1 (●F (Some $ to_agree n)) ∗ ticket_inv l2 γ2 ∗ (issued γ2 n ∗ ⌜b = true⌝ ∨ locked_at γ1 n ∗ ⌜b = false⌝).

  Global Program Instance newlock_spec :
    SPEC {{ True }}
      newlock #() 
    {{ (l1 l2 : loc) γ1 γ2, RET (#l1, #l2)%V; lock_state γ1 γ2 l1 l2 O false }}.

  Global Program Instance acquire_spec (l1 l2 : loc) γ1 γ2 :
    SPEC n b, << ▷ lock_state γ1 γ2 l1 l2 n b >>
      acquire (#l1, #l2)%V
    << RET #(); lock_state γ1 γ2 l1 l2 n true ∗ ⌜b = false⌝ ∗ locked_at γ1 n >>.

  Global Program Instance release_spec (l1 l2 : loc) γ1 γ2 (m : nat) :
    SPEC n b, << ▷ lock_state γ1 γ2 l1 l2 n b ∗ ▷ locked_at γ1 m >>
      release (#l1, #l2)%V
    << RET #(); lock_state γ1 γ2 l1 l2 (S n) false ∗ ⌜b = true⌝ ∗ ⌜n = m⌝ >>.
End proof.

Global Opaque lock_state.































