From iris.algebra Require Import coPset excl auth.
From diaframe.heap_lang Require Import proof_automation atomic_specs.
From iris.heap_lang Require Import proofmode notation atomic_heap.
From diaframe.lib Require Import ticket_cmra own_hints.
From diaframe.heap_lang.examples.logatom Require Import atomic_ticketlock.

Section proof.
  Context `{!heapGS Σ, !tlockG Σ} {aheap: atomic_heap}.

  Definition foo : val :=
    λ: "lk" "left" "right" "z",
      acquire "lk";;
      "left" <- "z";;
      "right" <- "z";;
      release "lk".

  Definition has_state γ1 γ2 lk1 lk2 l1 l2 (z : Z) : iProp Σ :=
      ∃ (n : nat) b, lock_state γ1 γ2 lk1 lk2 n b ∗
        (⌜b = true⌝ ∨ (l1 ↦ #z ∗ l2 ↦ #z ∗ ⌜b = false⌝)).
  (* the auth ghost state is necessary to remember that locked_at keeps the same value *)

  Obligation Tactic := program_smash_verify.

  Program Instance thread_spec_1 γ1 γ2 (lk1 lk2 l1 l2 : loc) (z : Z) :
    SPEC (m : Z), << ▷ has_state γ1 γ2 lk1 lk2 l1 l2 m >>
      foo (#lk1, #lk2)%V #l1 #l2 #z
    << RET #(); has_state γ1 γ2 lk1 lk2 l1 l2 z >>.
    (* Voila only verifies until here. We go one step further, using it in a client *)


  Definition run_foo : val :=
    λ: <>, 
      let: "lk" := newlock #() in
      let: "left" := ref #0 in
      let: "right" := ref #0 in
      Fork (foo "lk" "left" "right" #3);;
      Fork (foo "lk" "left" "right" #5);;
      acquire "lk";;
      if: ! "left" = ! "right" then 
        #()
      else 
        #() #() (* unsafe *).

  Global Program Instance thread_spec_2 γ1 γ2 (lk1 lk2 l1 l2 : loc) (z : Z) :
    SPEC {{ inv nroot (∃ z, has_state γ1 γ2 lk1 lk2 l1 l2 z) }}
      foo (#lk1, #lk2)%V #l1 #l2 #z
    {{ RET #(); True }}.
    (* thread_spec_1 shows the state transition is like we want. but this spec cannot be used
        in Fork (foo _) statements. So we derive a regular spec from the atomic one *)

  Global Program Instance run_foo_spec :
    SPEC {{ True }} run_foo #() {{ RET #(); True }}.
End proof.

















