From diaframe.heap_lang Require Import proof_automation atomic_specs wp_auto_lob.
From iris.algebra Require Import auth excl numbers.

From diaframe.heap_lang.examples.logatom Require Import queue_lib.


Definition new_queue : val := 
  λ: <>, let: "n" := AllocN #2 NONEV in AllocN #2 "n".

Definition dequeue : val :=
  rec: "dequeue" "q" :=
    let: "head" := ! "q" in
    let: "tail" := ! ("q" +ₗ #1) in
    let: "p" := NewProph in
    let: "next" := ! ("head" +ₗ #1) in

    let: "consistent":= "head" = ! "q" in
    resolve_proph: "p" to: "consistent";;
    if: "consistent" then
      if: "head" = "tail" then
        match: "next" with
          NONE => NONEV
        | SOME "l" => (* tail has fallen behind, help out *)
          CAS ("q" +ₗ #1) "tail" "l";;
          "dequeue" "q"
        end
      else
        match: "next" with
          NONE => #() #() (* error, should be unreachable *)
        | SOME "l" =>
          if: CAS "q" "head" "l" then 
            SOME ! "l"
          else 
            "dequeue" "q"
        end
    else "dequeue" "q".

Definition enqueue : val :=
  λ: "q" "v",
    let: "node" := AllocN #2 NONEV in
    let: "tailptr" := "q" +ₗ #1 in
    "node" <- "v" ;;
    (rec: "set_tail" <> :=
      let: "tail" := (! "tailptr") in
      let: "next" := ! ("tail" +ₗ #1) in
      if: ! "tailptr" = "tail" then
        match: "next" with
          NONE => 
            if: CAS ("tail" +ₗ #1) NONE (SOME "node") then 
              CAS "tailptr" "tail" "node";; #()
            else 
              "set_tail" #()
        | SOME "l" =>
            CAS "tailptr" "tail" "l";;
            "set_tail" #()
        end
      else "set_tail" #()
    ) #().

Definition queue_value_list_UR : ucmra := authUR $ max_prefix_listUR $ valO.
Definition queue_head_index_UR : ucmra := authUR $ max_natUR.
Class queueAltG Σ := QueueAltG { 
    queueG1 :> inG Σ queue_value_list_UR ;
    queueG2 :> inG Σ queue_head_index_UR;
    queueG3 :> inG Σ queue_base_UR ;
}.
Definition queueAltΣ : gFunctors := 
  #[GFunctor queue_value_list_UR; GFunctor queue_head_index_UR; GFunctor queue_base_UR].

Global Instance subG_queueAltΣ {Σ} : subG queueAltΣ Σ → queueAltG Σ.
Proof. solve_inG. Qed.


Section queue_atomic.
  Context `{!heapGS Σ, !queueAltG Σ}.
  Set Default Proof Using "Type*".

  Definition queue_head_inv γlv γll γih γit (l : loc) (ls : list loc) (vs : list val) (i : nat) : iProp Σ :=
    ∃ (l' : loc) (root : loc), 
      l ↦ #l' ∗ is_queue_base root ls vs ∗ own γlv (●{#1/2} to_max_prefix_list vs) ∗
        ⌜(root :: ls) !! i = Some l'⌝ ∗ own γih (●{#1/2} MaxNat i) ∗ own γit (◯ MaxNat i) ∗ own γll (● to_max_prefix_list (root :: ls)) ∗ ⌜length ls = length vs⌝.
        (* own γih (● MaxNat i) ∗ own γit (◯ MaxNat i) enforces that ● MaxNat i ≼ ● MaxNat j,
           which means that the tail pointer cannot lag further behind than the head pointer *)

  Definition is_queue γlv γih cvs : iProp Σ := ∃ (vs : list val) (i : nat), 
        own γlv (●{#1/2} to_max_prefix_list vs) ∗ own γih (●{#1/2} MaxNat i) ∗ ⌜cvs = drop i vs⌝.

  Definition queue_tail_inv (l : loc) γll γit : iProp Σ :=
    ∃ (l'' : loc) (ls : list loc) j, (l +ₗ 1) ↦ #l'' ∗ own γll (◯ to_max_prefix_list ls)
         ∗ ⌜ls !! j = Some l''⌝ ∗ own γit (● MaxNat j) ∗ ⌜NoDup ls⌝.

  Let queue_N := nroot.@"msc_queue".

  Definition is_queue_tail (l : loc) γll γit : iProp Σ := inv queue_N (queue_tail_inv l γll γit).
  Definition is_queue_head (l : loc) γlv γll γih γit := inv queue_N (∃ ls vs i, queue_head_inv γlv γll γih γit l ls vs i).
  Definition is_queue_inv (l : loc) γlv γll γih γit : iProp Σ := is_queue_head l γlv γll γih γit ∧ is_queue_tail l γll γit.

  (* lib.own_hints is not smart enough to figure these out, so we add them manually *)
  Instance max_nat_better γ n m :
    SolveSepSideCondition (n ≤ m) →
    HINT own γ (● MaxNat n) ✱ [- ; emp] ⊫ [bupd] ; own γ (● MaxNat m) ✱ [own γ (◯ MaxNat m)] | 15.
  Proof.
    rewrite /SolveSepSideCondition => Hnm; iStepsS.
  Qed.

  Instance core_id_auth_auth_remember γ {A : ucmra} dq (a : A) `{!inG Σ' $ authUR A} :
    CoreId a →
    HINT own γ (●{dq} a) ✱ [-; emp] ⊫ [bupd]; own γ (●{dq} a) ✱ [own γ (◯ a)] | 20.
  Proof.
    move => Ha; iStepS.
    iMod (own_update with "H1") as "HN".
    eapply auth_update_dfrac_alloc; last reflexivity. done.
    by iDestruct "HN" as "[$ $]".
  Qed.

  Global Instance new_queue_spec :
    SPEC {{ True }} 
      new_queue #() 
    {{ (l : loc) γll γlv γih γit, RET #l; 
      is_queue γlv γih [] ∗ is_queue_inv l γlv γll γih γit }}.
  Proof.
    iStepsS.
    iMod (own_alloc (● to_max_prefix_list ([]: list val))) as (?) "?"; first pure_solver.trySolvePure.
    iStepS.
    do 2 (iMod (own_alloc (● MaxNat 0)) as (?) "?"; first pure_solver.trySolvePure).
    iStepsS.
    iExists [],[], x. iStepsS.
  Qed.

  Hint Extern 4 (_ ≤ _)%nat => done : solve_pure_add.

  Definition will_be_consistent (xs : list (val * val)) : bool :=
    match xs with (* this kind of construction should be put in a specific format, so that we can add automatic rewriting for it. *)
    | [] => true  (* something like prophesize_to_resolve_to (val) *)
    | (v1, v2) :: xs =>
      bool_decide (v2 = #true)
    end.

  Global Instance dequeue_spec γll γlv γih γit (l : loc) :
    SPEC ⟨↑queue_N⟩ vs, << is_queue_inv l γlv γll γih γit ¦ is_queue γlv γih vs >>
      dequeue #l 
    << (ov : val), RET ov; ⌜ov = NONEV⌝ ∗ ⌜vs = []⌝ ∗ is_queue γlv γih []
                 ∨ ∃ v vs', ⌜vs = v :: vs'⌝ ∗ ⌜ov = SOMEV v⌝ ∗ is_queue γlv γih vs' >>.
  Proof.
    do 51 iStepS. (* stop after succesful load of next *)
    - (* next is NONE: linearize if head is consistent with our earlier load *)
      destruct (will_be_consistent x0) eqn:Hcons.
      * (* We need to prove: x11 = x13 for the if that occurs later *)
        assert (x15 = length x19); last subst.
        { apply lookup_lt_Some in H5. apply prefix_length in H11. simpl in H11.
          rewrite longest_length in H11.
          change (ofe_car locO) with loc in *. (* bloody hell *) lia. }
        assert (x12 = x23); last subst.
        { eapply prefix_cons_inv_1. by etrans. }
        assert (x8 = x19); last subst.
        { apply list_eq_from_prefix_length_ineq.
          + eapply prefix_cons_inv_2. etrans; done.
          + apply lookup_lt_Some in H1. simpl in H1. lia. }
        assert (x11 = x13); last subst.
        { enough (Some x11 = Some x13) by by simplify_eq. rewrite -H1.
          prove_lookup. }
        (* force comitting the linearization point *)
        iMod "H2". iReIntro "H2".
        iPoseProof (diaframe_close_right_quant with "H13") as "H13".
        iStepS. iLeft. iStepS. iStepS.
        iLeft. iStepS.
        (* finish, making sure to use our prophecy that head will be consistent *)
        do 23 iStepS. simpl in Hcons.
        apply bool_decide_eq_true_1 in Hcons. rewrite Hcons.
        iStepsS.
      * (* head will be inconsistent, do nothing and recurse *)
        iStepS. iLeft. do 20 iStepS.
        simpl in Hcons. apply bool_decide_eq_false_1 in Hcons.
        iRevert (Hcons). iStepS. rewrite H27.
        iStepsS.
    - (* next is SOME. skip case where head is inconsistent, then compare head with tail,
         inspect cases where a CAS was succesful *)
      do 47 iStepS; [ do 3 iStepS; last iStepsS | | iStepsS].
      * (* head = tail, but tail pointer is lagging, since next is SOME. 
          The following equalities also hold, but are irrelevant:   x27 = x10 and x15 = x10 *)
        assert (x28 = x10); last subst.
        { eapply (NoDup_lookup (longest x24 (x29 :: x25))). prove_nodup. prove_lookup. prove_lookup. }
        iStepS. split.
        { assert ((x29 :: x25) !! S x10 = Some x18); prove_lookup. }
        iStepS. iStepsS.
      * (* succesful dequeue of non empty queue. Open AU and commit *) 
        iMod "H2"; iReIntro "H2".
        iPoseProof (diaframe_close_right_quant with "H14") as "H14".
        iReIntro "H11".
        assert (x27 = x10); last subst.
        { eapply NoDup_lookup. exact H19. prove_lookup. prove_lookup. }
        assert (x30 = x10); last subst.
        { eapply NoDup_lookup. exact H27. prove_lookup. prove_lookup. }
        iStepS. unseal_diaframe => /=. iStepS. iStepS. iRight. iStepS.
        erewrite drop_S; last prove_lookup. solveSteps.
        iStepsS.
    Unshelve.
    + rewrite drop_ge //. lia.
    + rewrite drop_ge //. lia.
    + enough (x15 ≠ x10); first lia.
      contradict H26. subst.
      enough (Some x28 = Some x13) by by simplify_eq.
      rewrite -H21. prove_lookup.
  Qed.

  Global Instance enqueue_spec γll γlv γih γit (l : loc) (v : val) :
    SPEC ⟨↑queue_N⟩ vs, << is_queue_inv l γlv γll γih γit ¦ is_queue γlv γih vs >>
      enqueue #l v
    << RET #(); is_queue γlv γih (vs ++ [v]) >>.
  Proof.
    do 64 iStepS; [iLeft | ]; do 26 iStepS; last iStepsS. 
    (* allocates the node, performs Löb induction. Two cases: just after succesful help of tail pointer swing, and just before trying to enqueue. *)
    - do 8 iStepS; last iStepsS. (* if CAS fails, we recurse, this is easy *)
      iMod "H4"; iReIntro "H4". (* suppose CAS succeeds: need to commit AU *)
      iPoseProof (diaframe_close_right with "H14") as "H14".
      iReIntro "H9". iStepS. unseal_diaframe => /=. iStepS. iStepS.
      iStepS. iStepsS. (* failing swing is easy, iStepsS stops for succesful swing *)
      assert (x12 = x19 :: x1); last subst.
      { apply list_eq_from_prefix_length_ineq; first done.
        apply lookup_lt_Some in H0. simpl. lia. }
      assert (length x20 = length x1).
      { eapply NoDup_lookup. exact H25. prove_lookup. prove_lookup. }
      assert (x19 = x24); last subst.
      { eapply prefix_cons_inv_1. etrans; done. }
      assert (x20 = x1); last subst.
      { apply eq_sym, list_eq_from_prefix_length_ineq; last lia.
        apply (prefix_cons_inv_2 x24 x24). prove_prefix. }
      assert (x25 = length x1); last subst.
      { eapply (NoDup_lookup (longest x16 (x24 :: x1 ++ [x3]))). 
        prove_nodup. prove_lookup. prove_lookup. }
      split.
      { assert ((x24 :: x1 ++ [x3]) !! S (length x1) = Some x3); last prove_lookup.
        simpl. by apply list_lookup_middle. } (* now we have the needed equalities, we can finish *)
      iStepsS.
    - (* succesfully helped swing tail pointer forward *)
      iStepS. split.
      { eapply (prefix_lookup (x19 :: x15)). prove_lookup. prove_prefix. }
      iStepsS.
      enough (x13 = x25); first lia.
      eapply (NoDup_lookup (longest x24 (longest x22 (x19 :: x15)))).
      prove_nodup. prove_lookup. prove_lookup.
  Unshelve.
  + rewrite !app_length /=. lia.
  + rewrite drop_app_le //. apply lookup_lt_Some in H15. simpl in H15. lia.
  Qed.
End queue_atomic.

















