From diaframe.reloc Require Import proof_automation reloc_auto_lob backtracking.
From reloc Require Import examples.coinflip examples.lateearlychoice.

From diaframe.reloc.examples Require Import lateearlychoice.


Section coinflip_refinement.
  Context `{!relocG Σ, !lock.lockG Σ}.

(** Lazy coin (with prophecies) refines eager coin *)
  Definition I' (cl ce : loc) (p : proph_id) : iProp Σ :=
    (∃ vs : list (val*val), proph p vs ∗ ∃ v (b : bool), cl ↦ v ∗ ce ↦ₛ #b ∗
      (⌜v = NONEV⌝ ∗ ⌜b = extract_bool vs⌝
       ∨ ⌜v = SOMEV #b⌝)).

  Transparent lock.is_locked_r. (* TODO: bug, needed because newcoin and newlock have the same implementation *)

  Lemma coin_lazy'_eager_refinement_autom :
    ⊢ REL coin2' << coin1 : (() → lrel_bool) * (() → ()).
  Proof.
    unfold coin2', coin1. iStepsS.
    iAssert (|={⊤}=> ∃ γ, lock.is_lock coinN γ #x2 (I' x x4 x1))%I with "[-]" as ">HI";
    [ iStepsS | iReIntro "HI"].
    iSplitR; do 20 iStepS; unfold read, flip; repeat iStepR.
    (* execute left until lock is acquired, then execute right *)
  Qed.

  Opaque lock.is_locked_r.

  Lemma coin_eager_lazy'_refinement_autom :
    ⊢ REL coin1 << coin2' : (() → lrel_bool) * (() → ()).
  Proof.              (* this stuff   vvvv   is needed because of the bug described above *)
    unfold coin2', coin1. iStartProof2. rel_rec_l. iStepsS.
    iAssert (|={⊤}=> inv coinN (
        ∃ (b : bool), lock.is_locked_r x2 false ∗ x ↦ #b ∗ ∃ v, x0 ↦ₛ v ∗ (⌜v = NONEV⌝ ∨ ⌜v = SOMEV #b⌝)))%I with "[-]" as ">HI";
    [ iStepsS | iReIntro "HI"].
    iSplitR; iStepsS.
  Qed.

  Lemma coin_lazy'_lazy_refinement_autom :
    ⊢ REL coin2' << coin2 : (() → lrel_bool) * (() → ()).
  Proof.
    unfold coin2', coin2. iStepsS.
    iAssert (|={⊤}=> ∃ γ, lock.is_lock coinN γ #x2 (
        ∃ vs, proph x1 vs ∗ ∃ v, x ↦ v ∗ x3 ↦ₛ v ∗ (⌜v = NONEV⌝ ∨ ∃ (b : bool), ⌜v = SOMEV #b⌝)))%I with "[-]" as ">HI";
    [ iStepsS | iReIntro "HI"].
    iSplitR; do 20 iStepS; unfold read_lazy, flip_lazy; repeat iStepR; split; [ repeat f_equal | solveSteps; iStepsS].
    match goal with | |- ?b = ?x3 => unify b (extract_bool ((#(), #x3) :: x7)); done end.
  Qed.

  Lemma coin_lazy_lazy'_refinement_autom :
    ⊢ REL coin2 << coin2' : (() → lrel_bool) * (() → ()).
  Proof.
    unfold coin2', coin2. iStepsS.
    iAssert (|={⊤}=> inv coinN (
        lock.is_locked_r x2 false ∗ ∃ v, x ↦ v ∗ x0 ↦ₛ v ∗ (⌜v = NONEV⌝ ∨ ∃ (b : bool), ⌜v = SOMEV #b⌝)))%I with "[-]" as ">HI";
    [ iStepsS | iReIntro "HI"].
    iSplitR; iSmash.
  Qed.
End coinflip_refinement.