From iris.program_logic Require Import atomic.
From diaframe.reloc Require Import proof_automation symb_exec_logatom backtracking reloc_auto_lob.
From diaframe.heap_lang Require Import atomic_specs wp_auto_lob.
From reloc.lib Require Import lock.

(** Example from Fig. 2 in the paper. *)
Definition fg_incrementer : val :=
  λ: <>, 
    let: "l" := ref #1 in
    (rec: "f" <> :=
      let: "n" := ! "l" in
      if: CAS "l" "n" ("n" + #1) then
        "n"
      else
        "f" #()).

Definition cg_incrementer : val :=
  λ: <>,
    let: "l" := ref #1 in
    let: "lk" := newlock #() in
    (λ: <>,
      acquire "lk";;
      let: "n" := ! "l" in 
      "l" <- "n" + #1 ;;
      release "lk";;
      "n").

Section refinement.
Context `{!relocG Σ}.

Lemma fg_cg_incrementer_refinement :
  ⊢ REL fg_incrementer << cg_incrementer : () → () → lrel_int.
Proof.
  iStepsS.
  iAssert (|={⊤}=> inv (nroot.@"incr") 
      (∃ (n : nat), x ↦ #n ∗ x0 ↦ₛ #n ∗ is_locked_r x1 false))%I 
      with "[-]" as ">#Hinv"; first iStepsS.
  iSmash. (* try replacing [iSmash] with [iStepsS. iApply restore_mask. iStepsS.] *)
Qed.
End refinement.

(* above establishes logical refinement, contextual refinement follows: *)
Lemma fg_cg_incrementer_ctx_refinement :
  ∅ ⊨ fg_incrementer ≤ctx≤ cg_incrementer : () → () → TNat.
Proof.
  pose (Σ := #[relocΣ]).
  eapply (refines_sound Σ).
  iIntros (? Δ).
  iApply fg_cg_incrementer_refinement.
Qed.

(** ⊨ is notation for [ctx_refines]. We display its definition: *)
Print ctx_refines.
(** Note that this does not mention anything Diaframe/Iris specific! *)

(** Let us see the used axioms of this statement *)
Print Assumptions fg_cg_incrementer_ctx_refinement.
(** This shows two used axioms:
  - [refines_xchg_r], an Admitted lemma in the ReLoC source code. 
        Issue reported here https://gitlab.mpi-sws.org/iris/reloc/-/issues/11
        Proposed fix available here https://gitlab.mpi-sws.org/iris/reloc/-/merge_requests/6/diffs
  - Functional Extensionality, which is used throughout the Autosubst library, which is employed by ReLoC. *)


(** Example from Fig. 5 in the paper. *)
Definition inc : val := 
  rec: "f" "l" :=
    let: "n" := ! "l" in
    if: CAS "l" "n" ("n" + #1) then
      "n"
    else
      "f" "l".

Section inc_spec.
Context `{!heapGS Σ}.

Instance inc_spec (l : loc) :
  SPEC (z : Z), << l ↦ #z >> inc #l << RET #z; l ↦ #(z + 1) >>.
Proof. iSmash. Qed.

(** Above uses Diaframe's SPEC notation of logical atomicity.
  This does not trivially unfold to the Iris's logical atomicity. 
  We thus use it prove a logically atomic triple as defined by Iris. *)

Lemma inc_spec_iris (l : loc) :
  ⊢ atomic_wp (TB := [tele]) (inc #l)%E ∅ 
    (λᵗ (z : Z), l ↦ #z)%I
    (λᵗ z, tele_app (TT := [tele]) (l ↦ #(z + 1))%I)
    (λᵗ z, tele_app (TT := [tele]) #z).
  (* Parsing notation is broken, but it prints fine. *)
Proof. iSmash. Qed.
End inc_spec.

