From reloc Require Import examples.symbol lib.lock lib.list.
From diaframe.reloc Require Import proof_automation backtracking. (* backtracking not used *)
From diaframe.lib Require Import own_hints.
From iris.algebra Require Import auth numbers.

Section symbol_refinement.
  Context `{!relocG Σ, !msizeG Σ, !lock.lockG Σ}.

  Definition table_inv' γ (size1 size2 tbl1 tbl2 : loc) : iProp Σ :=
    (∃ (n : nat) (ls : val), size1 ↦{#1/2} #n ∗ size2 ↦ₛ{1/2} #n
                           ∗ own γ (● (MaxNat n))
                           ∗ tbl1 ↦{#1/2} ls ∗ tbl2 ↦ₛ{1/2} ls
                           ∗ lrel_list lrel_int ls ls)%I.

  Global Instance lrel_list_both_nil A :
    HINT ε₀ ✱ [- ; emp] ⊫ [id]; lrel_list A NILV NILV ✱ [emp].
  Proof. iStepS. rewrite -lrel_list_nil. iStepsS. Qed.

  Global Instance lrel_list_both_cons A v1 v2 ls1 ls2 :
    HINT ε₁ ✱ [- ; ▷ A v1 v2 ∗ ▷ lrel_list A ls1 ls2] ⊫ [id]; 
         lrel_list A (CONSV v1 ls1) (CONSV v2 ls2) ✱ [emp].
  Proof. iStepS. rewrite right_id. iApply lrel_list_cons; iStepsS. Qed.

  Hint Extern 4 (_ ≤ _)%nat => apply Nat.le_refl : solve_pure_add.

  Lemma refinement1' :
    ⊢ REL symbol1 << symbol2 : () → lrel_symbol.
  Proof.
    iStepsS.
    iAssert (|={⊤}=> ∃ γ γl, 
      inv sizeN (table_inv' γ x x2 x0 x3)
      ∗ is_lock (relocN.@"lock1") γl #x1 (lok_inv x x2 x0 x3 x4) )%I 
      with "[-]"
      as ">HN"; first by iStepsS. iReIntro "HN".
    iExists (tableR x5). iSplitR; [ | iSplitR]; iStepsS.
    - (* iSmash works here, but takes 4x as long. manually run store on RHS.*) do 44 iStepR. 
      iRestore. iStepsS. (* after store CONS: run all on RHS *) repeat iStepR.
    - (* after load, do load on RHS. *) repeat iStepR. rewrite bool_decide_true; last lia.
          do 8 iStepR. iRestore.
      iStepsS. (* after 2nd load, do 2nd load on RHS. *) repeat iStepR. iRestore.
      repeat iApply refines_app; [iApply nth_int_typed |iStepsS..].
  Qed.

  Lemma refinement2' :
    ⊢ REL symbol2 << symbol1 : () → lrel_symbol.
  Proof.
    iStepsS.
    iAssert (|={⊤}=> ∃ γ γl, 
      inv sizeN (table_inv' γ x x2 x0 x3) ∗
        is_lock (relocN.@"lock1") γl #x1 (lok_inv x x2 x0 x3 x4) )%I 
      with "[-]"
      as ">HN"; first by iStepsS. iReIntro "HN".
    iExists (tableR x5). iSplitR; [ | iSplitR]; iStepsS.
    - (* iSmash works here, but takes 4x as long. manually run store on RHS.*) do 44 iStepR.
      iRestore. iStepsS. (* after store CONS: run all on RHS *) repeat iStepR.
    - rewrite bool_decide_true; last lia.
      iStepsS. (* first op: do nothing on RHS. *) iRestore.
      iStepsS. (* this load: do load on RHS. *) do 8 iStepR. iRestore.
      iStepsS. (* 2nd load: do 2nd load on RHS. *) repeat iStepR. iRestore.
      repeat iApply refines_app; [iApply nth_int_typed |iStepsS..].
  Qed.
End symbol_refinement.