From reloc Require Import examples.various lib.lock lib.counter.
From diaframe.reloc Require Import proof_automation counter backtracking.
From diaframe.lib Require Import own_hints.
From iris.algebra Require Import cmra numbers agree excl.

Section proofs.
  Context `{!relocG Σ}.

  Instance update_shot γ `{oneshotG Σ} :
    HINT pending γ ✱ [- ; emp] ⊫ [bupd]; shot γ ✱ [shot γ].
  Proof. iStepS. rewrite shoot. iMod "H1" as "#H". iStepsS. Qed.

  Lemma refinement2' `{oneshotG Σ} :
    ⊢ REL
      let: "x" := ref #0 in
      (λ: "f", "x" <- #1;; "f" #();; !"x")
    <<
      (let: "x" := ref #1 in
       λ: "f", "f" #();; !"x")
    : (() → ()) → lrel_int.
  Proof.
    iStepsS.
    iAssert (|={⊤}=> ∃ γ, inv shootN (∃ (n : nat), 
          x ↦ #n ∗ (⌜n = 0⌝ ∗ pending γ  ∨ ⌜n = 1⌝ ∗ shot γ) ∗ 
          x0 ↦ₛ #1))%I
         with "[-]"
         as ">[%γ #Hinv]"; first by iStepsS.
    iSmash.
  Qed.

  Lemma refinement25' `{oneshotG Σ} :
    ⊢ REL
      (λ: "f", "f" #();; #1)
    <<
      (let: "x" := ref #0 in
       (λ: "f", "x" <- #1;; "f" #();; !"x"))
    : (() → ()) → lrel_int.
  Proof. 
    iStepsS.
    iAssert (|={⊤}=> ∃ γ, inv shootN (∃ (n : nat), 
          x ↦ₛ #n ∗ (⌜n = 0⌝ ∗ pending γ  ∨ ⌜n = 1⌝ ∗ shot γ)))%I
         with "[-]"
         as ">[%γ #Hinv]"; first by iStepsS.
    iStepsS; repeat iStepR.
  Qed.

  Definition i3' (x x' b b' : loc) : iProp Σ :=
    (∃ (n:nat), x ↦{#1/2} #n ∗ x' ↦ₛ{1/2} #n ∗
     ∃ (bv:bool), b ↦ #bv ∗ b' ↦ₛ #bv ∗
    ((⌜bv = true⌝ ∗ x ↦{#1/2} #n ∗ x' ↦ₛ{1/2} #n)
    ∨ (⌜bv = false⌝)))%I.

  Lemma refinement3' :
    ⊢ REL
      let: "b" := ref #true in
      let: "x" := ref #0 in
      (λ: "f", if: CAS "b" #true #false
               then "f" #();; "x" <- !"x" + #1 ;; "b" <- #true
               else #(),
       λ: <>, !"x")
    <<
      (let: "b" := ref #true in
      let: "x" := ref #0 in
      (λ: "f", if: CAS "b" #true #false
               then let: "n" := !"x" in
                    "f" #();; "x" <- "n" + #1 ;; "b" <- #true
               else #(),
       λ: <>, !"x"))
    : ((() → ()) → ()) * (() → lrel_int).
  Proof.
    iStepsS.
    iMod (inv_alloc i3n _ (i3' x0 x2 x x1) with "[-]") as "#Hinv"; first iStepsS.
    iSplitR; iSmash.
  Qed.

  Lemma refinement4' `{!lockG Σ}:
    ⊢ REL
      (let: "x" := ref #1 in
       let: "l" := newlock #() in
       λ: "f", acquire "l";;
               "x" <- #0;; "f" #();;
               "x" <- #1;; "f" #();;
               let: "v" := !"x" in
               release "l";; "v")
    <<
      (let: "x" := ref #0 in
       λ: "f", "f" #();; "x" <- #1;; "f" #();; !"x")
    : (() → ()) → lrel_int.
  Proof.
    iStepsS.
    pose (N:=relocN.@"lock").
    iAssert (|={⊤}=> ∃ γ, is_lock N γ #x0 (∃ (n m : Z), x ↦ #n ∗ x1 ↦ₛ #m))%I 
      with "[-]" as ">[%γ #HN]"; first iStepsS.
    iStepsS. repeat iStepR.
  Qed.

  Lemma refinement5' :
    ⊢ REL
      (λ: "f", let: "x" := ref #0 in
               let: "y" := ref #0 in
               "f" #();;
               "x" <- !"y";;
               "y" <- #1;;
               !"x")
    <<
      (λ: "f", let: "x" := ref #0 in
               let: "y" := ref #0 in
               "f" #();;
               "x" <- !"y";;
               "y" <- #2;;
               !"x")
    : (() → ()) → lrel_int.
  Proof. repeat iStepR. Qed.

  Definition i6' `{oneshotG Σ} `{inG Σ (exclR unitR)} (c1 c2 : loc) γ γ' :=
    (∃ (n n' : nat), 
      c1 ↦ #n' ∗ c2 ↦ₛ #n ∗
     ((⌜n' = n⌝ ∗ pending γ)
   ∨ (⌜n' = (n + 1)%nat⌝ ∗ shot γ ∗ own γ' (Excl ()))))%I.

  Lemma refinement6' `{oneshotG Σ} `{inG Σ (exclR unitR)} :
    ⊢ REL
      (λ: "f" "g" "f'",
       let: "pg" := p "g" in
       let: "g'" := Fst "pg" in
       let: "g''" := Snd "pg" in
       "f" "g'";; "g'" #();; "f'" "g'";; "g''" #())
    <<
      (λ: "f" "g" "f'",
       let: "pg" := p "g" in
       let: "g'" := Fst "pg" in
       let: "g''" := Snd "pg" in
       "f" "g'";; "g" #();; "f'" "g'";; "g''" #() + #1)
    : ((() → ()) → ()) → (() → ()) → ((() → ()) → ()) → lrel_int.
  Proof.
    repeat iStepR.
    iAssert (|={⊤}=> ∃ γ γ', own γ' (Excl ()) ∗ inv shootN (i6' x6 x5 γ γ'))%I 
        with "[H4 H5]" as ">[%γ [%γ' [He #Hinv]]]"; first iStepsS.
    iSplitR "He"; [iSmash | iStepsS].
    repeat iStepR; iSplitR; iSmash.
  Qed.

End proofs.