From diaframe Require Import proofmode_base.
From iris.bi Require Import bi.
From iris.base_logic Require Import lib.invariants.
From iris.proofmode Require Import proofmode.
From diaframe.lib Require Import iris_hints.

Section later_problems.
  Context {PROP : bi}.
  Implicit Types P Q G : PROP.

  Lemma test_local_later P Q G :
    (P -∗ Q) ∗ ▷ P ∗ G ⊢ ▷ Q ∗ G.
  Proof.
    iStepS.
    iStepsS. (* Problem: Have ▷ A, need to prove A since Diaframe eagerly introduces ▷ on B *)
    Undo 1.
    iFrame. (* Now G gets removed, so we will introduce the later and everything is fine *)
    iStepsS.
  Qed.

  Lemma test_local_later_hard {A : Type} (P Q R : A → PROP) c G :
    Affine (PROP := PROP) True%I → (∀ a, Affine (R a)) →
    ((∀ a, P a -∗ (Q a ∗ R a)) ∗ ▷ P c ∗ G ⊢ (∃ b, ▷ Q b) ∗ G).
  Proof.
    move => HTrue HR.
    iStepS. 
    iStepsS. 
    Undo 1.
    iFrame.
    iStepsS. (* Still a problem now! Works if the ▷ stays outside the existential *)
    Undo 1.
    assert (Inhabited A) by exact (populate c).
    erewrite <-bi.later_exist.
    iStepsS.
  Qed.

  Lemma test3_local_later_hard_abd {A : Type} (P Q : A → PROP) c : ▷ P c ∗ (∀ a, P a -∗ Q a) ⊢ ∃ a, ▷ Q a.
  Proof.
    iStepS.
    iStepS. (* abduct could be relatively easily fixed with the new abduct stuff, but BiAbd is not ready.. *)
    Undo 1. 
    iSpecialize ("H2" with "H1").
    iStepsS. 
  Qed.
End later_problems.

Section fupd_problems.
  Context `{!invGS Σ}.

  Variable P Q R : iProp Σ.

  Variable HR : HINT ε₁ ✱ [- ; R] ⊫ [id]; P -∗ Q ✱ [emp].

  Lemma test_inv_biabd :
    inv nroot (P -∗ Q) ∗ P ∗ R ⊢ |={⊤}=> ▷ Q ∗ emp.
  Proof. iStepsS. Qed.

  Lemma test_inv_abd :
    inv nroot (P -∗ Q) ∗ P ∗ R ⊢ |={⊤}=> ▷ Q.
  Proof. iStepsS. Qed.
    (* now automatic! :) *)

  Lemma test_wand_mod1 (E1 E2 : coPset) :
    (P -∗ |={E1, E2}=> Q) ∗ (True -∗ P ∗ |={E2, E1}=> emp) ⊢ |={E1}=> Q.
  Proof. iStepsS. Qed.

  Lemma test_wand_mod2 (E1 E2 : coPset) :
    (P -∗ |={E1, E2}=> Q) ∗ (|={E1}=> P ∗ |={E2, E1}=> emp) ⊢ |={E1}=> Q.
  Proof. iStepsS. Qed.

  Lemma test_wand_mod3 (E1 E2 E3 : coPset) :
    (P -∗ |={E3, E2}=> Q) ∗ (|={E1, E3}=> P ∗ |={E2, E1}=> emp) ⊢ |={E1}=> Q.
  Proof. iStepsS. Qed.

  Lemma test_forall_mod1 E1 E2 `(S : A → iProp Σ) :
    (∀ a, S a -∗ |={E1, E2}=> Q) ∗ (|={E1}=> ∃ x, S x ∗ |={E2,E1}=> emp) ⊢ |={E1}=> Q.
  Proof. iStepsS. Qed.

  Lemma test_forall_mod_equiv E1 E2 `(S : A → iProp Σ) :
    ((∃ a, S a) -∗ |={E1, E2}=> Q) ∗ (|={E1}=> ∃ x, S x ∗ |={E2,E1}=> emp) ⊢ |={E1}=> Q.
  Proof. iStepsS. Qed.
End fupd_problems.