From diaframe.heap_lang Require Import proof_automation wp_auto_lob.
From iris.heap_lang Require Import proofmode.

(* This file can be used to test the overhead of Diaframe's automation on relatively simple goals.
   For the below test, proofs are about 6x slower than a tailored tactic. *)

Ltac Zify.zify_post_hook ::= Z.div_mod_to_equations.

Section stress_test1.
  Context `{!heapGS Σ}.

  Definition sum_rec : val :=
    rec: "sum_until" "l" "x" "i_max" :=
      if: ("x" ≤ "i_max") then
        "l" <- !"l" + "x" ;; 
        "sum_until" "l" ("x" + #1) "i_max"
      else #().

  Definition sum_until : val :=
    λ: "i_max",
      let: "l" := ref #0 in
      sum_rec "l" #1 "i_max" ;; 
      ! "l".

  Proposition sum_until_20_direct : {{{ True }}} sum_until #20 {{{ RET #210; True }}}.
  Proof.
    iStepsS.
    time repeat (try (iIntros "!>"); wp_lam; iStepsS).
  Time Qed.

  Instance sum_rec_spec (l : loc) (n m s : Z) : 
    SPEC {{ l ↦ #s }}
      sum_rec #l #n #m 
    {{ (z : Z), RET #(); l ↦ #z ∗ (⌜z = (s + ((n + m - 1) * (m - n) + 2 * m) / 2)%Z⌝ ∗ ⌜n ≤ m⌝%Z ∨ ⌜z = s⌝ ∗ ⌜m < n⌝%Z) }}.
  Proof.
    do 47 iStepS. (* TODO: Prevent unification of evars in postconditions *)
    iApply (wp_wand with "[H2]"); iStepsS.
    assert (x2 = x3) as ->; lia.
  Qed.

  Instance sum_until_spec (n : Z) : 
    SPEC {{ ⌜(0 ≤ n)%Z⌝ }} sum_until #n {{ RET #(n * (n + 1) / 2)%Z; True }}.
  Proof. iStepsS. Qed.

  Proposition sum_until_20_from_spec : {{{ True }}} sum_until #20 {{{ RET #210; True }}}.
  Proof. iStepsS. Qed.

End stress_test1.


Section stress_test_old.
  Context `{!heapGS Σ}.

  Proposition sum_until_20_old : {{{ True }}} sum_until #20 {{{ RET #210; True }}}.
  Proof.
    iIntros (Φ) "_ Post".
    wp_rec.
    wp_alloc l as "Hl".
    do 2 wp_pure _.
    wp_lam.
    time "Old implementation of wp_tacs" repeat (
      wp_load; wp_store; wp_pures; wp_lam); wp_load.
    by iApply "Post".
  Time Qed.

End stress_test_old.



















